package com.hipme.flipsheet;

import danger.app.*;
import danger.internal.Date;
import danger.ui.*;
import danger.util.LinkedList;
import danger.util.ListIterator;
import danger.util.StringTokenizer;
import danger.system.*;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;
import danger.util.DEBUG;
import danger.util.format.StringFormat;

import com.hipme.flipsheet.parser.Pratt;

public final class SheetTable extends Table implements Resources {
  static final int maxRows = 64;
  static final int nRows = maxRows;
  static final int[] cellWidths = { 12, 55, 55, 55, 60};
  static final int nCols = cellWidths.length;
  private static int HEADER_BACKGROUND_COLOR = Style.getSystemStyle().getColor(Style.TITLE_BAR_PILL_COLOR);
  private static int HEADER_FOREGROUND_COLOR = Style.getSystemStyle().getColor(Style.TITLE_BAR_TEXT_COLOR);

  int cellHeight;
  Font valueFont;
  Font labelFont;

  String sheetName;
  int rowCursor = 1;
  int colCursor = 1;

  static int defaultBackgroundColor = Color.WHITE;
  static int highlightBackgroundColor = Color.YELLOW;
  boolean wheelAxisHorizontal = true;
  Pratt parser;
  LinkedList masterDependencies;
  LinkedList onLoad=null;
  Rect tmpRect = new Rect();	// good idea or not?
  /**
   * If true, take descenders into account to "g" and "p" look nice,
   * lose a couple of lines of the screen.
   */
  boolean showDescenders = false;
  
  /**
   * If true, outline rather than use background.
   */
  boolean useOutline = false;

  public SheetTable(String csv) {
    this(csv, getDefaultSheetName(((flipsheet)Application.getCurrentApp()).getSheetNumber()));
  }

  public SheetTable(String csv, String sheetName) {
    super();
    initialize(csv, sheetName);
  }

  int getRowCursor() {return rowCursor;}
  int getColCursor() {return colCursor;}

  static String getDefaultSheetName(int i) {
    return StringFormat.fromRsrc(kStrSheetFormat, String.valueOf(i));
  }

  String getSheetName() {
    return sheetName;
  }

  void setSheetName(String sheetName) {
    this.sheetName=sheetName;
  }

  String getCellName() {
    // cache this?
    return ((char)('A' + (colCursor - 1))) + "" + (rowCursor);
  }

  String getOnLoadText() {
    if (onLoad == null) return "";
    StringBuffer sb = new StringBuffer();
    for (ListIterator it = onLoad.listIterator(); it.hasNext(); ) {
      sb.append(it.next());
      sb.append("\n");
    }
    return sb.toString();
  }

  void setOnLoadText(String text) {
    onLoad = new LinkedList();
    addOnLoadText(text);
  }

  /**
   * For now, all this does is append a series of expressions.
   * Some day it should merge ranges to eliminate duplicates
   * and flip/flops.
   */
  void addOnLoadText(String text) {
    StringTokenizer st = new StringTokenizer(text, "\r\n");
    while (st.hasMoreTokens()) {
      addOnLoadExpression(st.nextToken());
    }
  }

  /**
   * For now, all this does is append one expression
   * Some day it should merge ranges to eliminate duplicates
   * and flip/flops.
   */
  void addOnLoadExpression(String expr) {
    try {
      String value = evaluate(expr);
      DEBUG.p(expr + "->" + value);
      onLoad.addLast(expr);
    } catch (Exception e) {
      DEBUG.p("onLoad expr " + expr + " got error " + e);
      e.printStackTrace();
    }
  }

  void nextRow() { 
    setRowCursor(rowCursor+1); 
  }

  void nextCol() { 
    setColCursor(colCursor+1); 
  }

  void previousRow() { 
    setRowCursor(rowCursor-1); 
  }

  void previousCol() { 
    setColCursor(colCursor-1); 
  }

  /**
   * Moves forward one cell.  Wraps to previous or next row
   * at ends of line.  Returns true if wrapped, false if not.
   */
  boolean forwardCell() {
    if (colCursor == nCols-1) {
      setCursor(rowCursor+1, 1);
      return true;
    } else {
      nextCol();
      return false;
    }
  }

  /**
   * Moves back one cell.  Wraps to previous or next row
   * at ends of line.  Returns true if wrapped, false if not.
   */
  boolean backCell() {
    if (colCursor == 1) {
      setCursor(rowCursor-1, nCols-1);
      return true;
    } else {
      previousCol();
      return false;
    }
  }

  void setCursor(int row, int col) {
    if (row < 1 || col < 1) return;
    if (row >= getRows() || col >= getColumns()) return;
    clearHighlighted(rowCursor, colCursor);
    rowCursor = row;
    colCursor = col;
    setHighlighted(rowCursor, colCursor);
  }

  void setRowCursor(int i) {
    if (i < 1) return;
    if (i >= getRows()) return;
    clearHighlighted(rowCursor, colCursor);
    rowCursor = i;
    setHighlighted(rowCursor, colCursor);
  }

  void setColCursor(int i) {
    if (i < 1) return;
    if (i >= getColumns()) return;
    clearHighlighted(rowCursor, colCursor);
    colCursor = i;
    setHighlighted(rowCursor, colCursor);
  }

  void clearHighlighted(int row, int col) {
    SheetTableCell td = ((SheetTableCell)getTableCell(row, col));
    if (td != null) {
      td.setHighlighted(false);
      getCellBounds(row, col, tmpRect);
      if (useOutline) includeCellBorder(tmpRect);
      invalidate(tmpRect);
    } else {
      // DEBUG.p("setCellBackgroundColor " + row +","+col+ " was null");
    }
  }

  void setHighlighted(int row, int col) {
    SheetTableCell td = ((SheetTableCell)getTableCell(row, col));
    if (td != null) {
      td.setHighlighted(true);
      getCellBounds(row, col, tmpRect);
      if (useOutline) includeCellBorder(tmpRect);
      invalidate(tmpRect);
    } else {
      // DEBUG.p("setCellBackgroundColor " + row +","+col+ " was null");
    }
  }

  public void setCellBackgroundColor(int row, int col, int color) {
    SheetTableCell td = (SheetTableCell)getTableCell(row, col);
    if (td != null) {
      td.backgroundColor=color;
      td.setBackgroundColor(color);
      getCellBounds(row, col, tmpRect);
      invalidate(tmpRect);
    } else {
      // DEBUG.p("setCellBackgroundColor " + row +","+col+ " was null");
    }
  }

  public void setCellForegroundColor(int row, int col, int color) {
    SheetTableCell td = (SheetTableCell)getTableCell(row, col);
    if (td != null) {
      td.foregroundColor=color;
      td.setForegroundColor(color);
      getCellBounds(row, col, tmpRect);
      invalidate(tmpRect);
    } else {
      // DEBUG.p("setCellForegroundColor " + row +","+col+ " was null");
    }
  }

  Rect getFastCellBounds(Rect rect) {
    return getFastCellBounds(rowCursor, colCursor, rect);
  }

  /**
   * Get only the Y portion of the bounds and leave the X as table bounds,
   * for use with ScrollView, which only cares about the Y.
   */
  Rect getFastCellBounds(int row, int col, Rect rect) {
    // fill in to the size of the table to get the X.
    getBounds(rect);
    int borderWidth = getBorderWidth();
    int prevBot = borderWidth;
    for (int j = 0; j <= row; j++) {
      int height = getTableCell(j, 0).getHeight();
      rect.top = prevBot;
      rect.bottom = rect.top + height;
      prevBot = prevBot + height + borderWidth;
    }
    return rect;
  }

  /**
   * This method returns the cell bounds accurate
   * in both X and Y directions, but is more expensive
   * to compute.
   */
  Rect getCellBounds(int row, int col, Rect rect) {
    getBounds(rect);
    int borderWidth = getBorderWidth();
    {
      int prevBot = borderWidth;
      for (int j = 0; j <= row; j++) {
	int height = getTableCell(j, 0).getHeight();
	rect.top = prevBot;
	rect.bottom = rect.top + height;
	prevBot = prevBot + height + borderWidth;
      }
    }
    {
      int prevLeft = borderWidth;
      for (int i = 0; i <= col; i++) {
	int width = getTableCell(row, i).getWidth();
	rect.left = prevLeft;
	rect.right = rect.left + width;
	prevLeft = prevLeft + width + borderWidth;
      }
    }
    return rect;
  }

  Rect getCellBounds(Rect rect) {
    return getCellBounds(rowCursor, colCursor, rect);
  }

  Rect includeCellBorder(Rect rect) {
    rect.top -= 1;
    rect.left -= 1;
    rect.right += 2;
    rect.bottom += 2;
    return rect;
  }


  /**
   * ScrollView will sometimes scroll to make sure this rect is in view,
   * so we might as well use it ourselves.
   */
  public Rect getInterestingRect(Rect rect) {
    return getFastCellBounds(rect);
  }

  void setHighlightBackgroundColor(int color) {
    highlightBackgroundColor = color;
  }

  boolean isWheelAxisHorizontal() {
    return wheelAxisHorizontal;
  }

  void setWheelAxisHorizontal(boolean w) {
    wheelAxisHorizontal = w;
    setCellContents(0, 0, wheelAxisHorizontal ? "-" : "|");
    getCellBounds(0, 0, tmpRect);
    invalidate(tmpRect);
  }

  public boolean addRow(Vector vector) {
    boolean value = super.addRow(vector);
    if (getRows() == rowCursor - 1)
      setCellBackgroundColor(getRowCursor(), getColCursor(), highlightBackgroundColor);
    return value;
  }

  TableCell getCell() {
    return getTableCell(getRowCursor(), getColCursor());
  }

  String getCellContents() {
    return getCellContents(getRowCursor(), getColCursor());
  }

  String getCellContents(int row, int col) {
    TableCell td = getTableCell(row, col);
    if (td == null) return null;
    if (td instanceof SheetTableCell) {
      return ((SheetTableCell)td).getContents();
    } else {
      return td.getElementAsString();
    }
  }

  void setCellContents(String value) {
    setCellContents(getRowCursor(), getColCursor(), value);
  }

  void setCellContents(int row, int col, String newValue) {
    TableCell td = getTableCell(row, col);
    if (td instanceof SheetTableCell) {
      ((SheetTableCell)td).setCellContents(newValue);
    } else {
      td.setElementToString(newValue);
    }
  }

  boolean isCellLocked() {
    return ((SheetTableCell)getCell()).isLocked();
  }

  boolean isCellLocked(int row, int col) {
    TableCell td = getTableCell(row, col);
    if (td instanceof SheetTableCell) {
      return ((SheetTableCell)td).isLocked();
    } else {
      return true;
    }
  }

  public void setLocked(boolean locked) {
    ((SheetTableCell)getCell()).setLocked(locked);
  }

  public void setLocked(int row, int col, boolean locked) {
    ((SheetTableCell)getTableCell(row, col)).setLocked(locked);
  }

  /**
   * Normally we show descenders like "y" but to get 
   * several more rows on the screen you can turn that off
   */
  void setShowDescenders(boolean showDescenders) {
    //DEBUG.p("showDescenders = " + showDescenders);
    //DEBUG.p("this.showDescenders = " + this.showDescenders);
    if (this.showDescenders == showDescenders) return;
    this.showDescenders = showDescenders;

    calculateCellHeight();
    for (int row = 0; row < nRows; row++) {
      for (int col = 0; col < nCols; col++) {
	getTableCell(row, col).setSize(cellWidths[col], cellHeight);
      }
    }
    invalidate();
  }

  boolean getShowDescenders() {
    return showDescenders;
  }

  /**
   * Use outline or highlight background color fur current cell.
   */
  void setUseOutline(boolean useOutline) {
    if (this.useOutline == useOutline) return;
    clearHighlighted(rowCursor, colCursor);
    this.useOutline = useOutline;
    setHighlighted(rowCursor, colCursor);
  }

  boolean getUseOutline() {
    return useOutline;
  }

  void calculate() {
    if (masterDependencies == null) {
      long t = System.currentTimeMillis();
      masterDependencies = calculateMasterDependencies();
      DEBUG.p("Calculating masterDependencies took " + (System.currentTimeMillis()-t) + " ms");
      // DEBUG.p("MasterDependencies = " + plist(masterDependencies));
    }
    if (masterDependencies == null) {
      DEBUG.p("Skipping calculations because of circular dependencies");
      return;
    }
    {
      long t = System.currentTimeMillis();
      for (ListIterator it = masterDependencies.listIterator(); it.hasNext(); ) {
	SheetTableCell td = (SheetTableCell)it.next();
	td.calculate();
      }
      DEBUG.p("Calculating only required values took " + (System.currentTimeMillis()-t) + " ms");
    }
  }

  void invalidateMasterDependencies() {
    masterDependencies = null;
  }

  LinkedList calculateMasterDependencies() {
    LinkedList q = new LinkedList();
    LinkedList masterDependencies = new LinkedList();
    Hashtable degrees = new Hashtable();
    int num_vertex = 0;
    // Add all cells with inDegree=0 to the queue.
    for (int j = 1; j < nRows; j++) {
      for (int i = 1; i < nCols; i++) {
	SheetTableCell v = (SheetTableCell)getTableCell(j, i);
	if (v.getFormula() != null || v.getDependents() != null) {
	  num_vertex++;
	  LinkedList dependencies = v.getDependencies();
	  if (dependencies == null || dependencies.size() == 0) {
	    q.addLast(v);
	  } else {
	    degrees.put(v, new Integer(dependencies.size()));
	  }
	}
      }
    }
    while (q.size() > 0) {
      SheetTableCell v = (SheetTableCell)(q.removeFirst());
      masterDependencies.addLast(v);
      LinkedList dependents = v.getDependents();
      if (dependents != null) {
	ListIterator it = dependents.listIterator();

	while (it.hasNext()) {
	  SheetTableCell w = ((SheetTableCell)it.next());
	  Integer d1 = (Integer)degrees.get(w);
	  if (d1 != null) {
	    int deg = d1.intValue()-1;
	    if (deg == 0) {
	      degrees.remove(w);	// double check
	      q.addLast(w);
	    } else {
	      degrees.put(w, new Integer(deg));
	    }
	  } else {
	    DEBUG.p("For cell " + v + " the dependency " + w + " was not in the table");
	  }
	}
      }
    }
    if (masterDependencies.size() > num_vertex) {
      DEBUG.p("circular dependency: masterDependencies.size = " + masterDependencies.size() +
	      " num_vertex = " + num_vertex);
    }
    if (true) {
      if (degrees.size() != 0) {
	java.util.Enumeration e = degrees.keys();
	StringBuffer sb = new StringBuffer("Circular references: "); // I18N
	while (e.hasMoreElements()) {
	  SheetTableCell k = (SheetTableCell)(e.nextElement());
	  sb.append(k.toString());
	  sb.append(" ");
	  // not quite right because current cell bg cleared before cursor moves on.
	  k.setBackgroundColor(Color.RED);
	  invalidate(getCellBounds(k.row, k.col, tmpRect));
	}
	((flipsheet)Application.getCurrentApp()).showCircularDependency(sb.toString());
	masterDependencies = null;
      }
      if (q.size() != 0)
	throw new RuntimeException("BUG: q.size != 0: " + degrees.size());
    }
    return masterDependencies;
  }

  static String plist(LinkedList list) {
    if (list == null) return "";
    StringBuffer sb = new StringBuffer();
    for (ListIterator it = list.listIterator(); it.hasNext(); ) {
      sb.append(it.next());
      sb.append(" ");
    }
    return (sb.toString());
  }

  void clear() {
    for (int col = 1; col < nCols; col++) {
      for (int row = 1; row < nRows; row++) {
	SheetTableCell td = (SheetTableCell)getTableCell(row, col);
	td.setCellContents(null);
	td.setForegroundColor(Color.BLACK);
	td.setBackgroundColor(Color.WHITE);
      }
    }
    setWheelAxisHorizontal(true);
    setCursor(1, 1);
    onLoad=null;
    invalidate();		// necessary?
  }

  void initialize(String csv, String sheetName) {
    // DEBUG.p("initialize " + sheetName);
    Vector rowVec;
    final int borderWidth = 1;
    final int cellSpacing = 1;

    this.sheetName = sheetName;

    useOutline = ((flipsheet)Application.getCurrentApp()).useOutline;
    // Font.findFont("Fixed6x10");
    valueFont = Font.findSystemFont();
    labelFont = Font.findBoldSystemFont();
    calculateCellHeight();
    setPosition(0, 0);
    setLeft(0);
    setBorderWidth(borderWidth);
    setCellSpacing(cellSpacing);
    setBackgroundColor(defaultBackgroundColor);
    setBorderColor(Color.GRAY11);
    setSpacingColor(Color.GRAY11);

    // Make column headings in row 0
    {
      rowVec = new Vector();
      for (int col = 0; col < nCols; col++) {
	TableCell td = new TableCell(this);
	td.setSize(cellWidths[col], cellHeight);
	td.setForegroundColor(HEADER_FOREGROUND_COLOR);
	td.setBackgroundColor(HEADER_BACKGROUND_COLOR);
	if (col == 0) {
	  td.setElementToString("");
	} else {
	  td.setElementToString("" + ((char)('A' + col - 1)));
	}
	td.setFont(labelFont);
	td.setAlignment(StaticText.HALIGN_CENTER);
	rowVec.addElement(td);
      }
      addRow(rowVec);
    }

    // Make row headings and fill table 
    for (int row = 1; row < nRows; row++) {
      rowVec = new Vector();
      for (int col = 0; col < nCols; col++) {
	if (col == 0) {
	  TableCell td = new TableCell(this);
	  td.setSize(cellWidths[col], cellHeight);
	  td.setForegroundColor(HEADER_FOREGROUND_COLOR);
	  td.setBackgroundColor(HEADER_BACKGROUND_COLOR);
	  td.setElementToInt(row);
	  td.setFont(labelFont);
	  td.setAlignment(StaticText.HALIGN_RIGHT);
	  rowVec.addElement(td);
	} else {
	  // Really should make these sparse instead.
	  SheetTableCell td = new SheetTableCell(this, row, col);
	  td.setSize(cellWidths[col], cellHeight);
	  td.setForegroundColor(Color.BLACK);
	  td.setFont(valueFont);
	  rowVec.addElement(td);
	}
      }
      addRow(rowVec);
    }

    // Autosize widget
    autoSize();

    // if CSV is available, fill table now that
    // all the cells exist.
    CSVParser parser = null;
    if (csv != null && csv.length() > 0) {
      long t = System.currentTimeMillis();
      parser = new CSVParser(csv);
      // DEBUG.p("setting precision to " + parser.precision);
      setPrecision(parser.precision);
      boolean done = false;
      for (int row = 1; row < nRows && !done; row++) {
	for (int col = 1; col < nCols; col++) {
	  if (!parser.hasNext()) {
	    done = true;
	    break;
	  }
	  String v = (String)parser.next();
	  SheetTableCell td = (SheetTableCell)getTableCell(row, col);
	  td.setCellContents(v);
	}
      }
      DEBUG.p("Setting values from CSV took " + (System.currentTimeMillis() - t) + "ms");
      // Check for onLoad
      onLoad=parser.onLoad;
      if (onLoad != null) {
	for (ListIterator it = onLoad.listIterator(); it.hasNext(); ) {
	  String expr = (String)it.next();
	  try {
	    DEBUG.p(expr + "->" + evaluate(expr));
	  } catch (Exception e) {
	    DEBUG.p("onload expr = " + expr + " got error " + e);
	  }
	}
     } 
    } else {
      // no CSV, nothing to do.
    }

    // Restore or initialize table parameters
    if (parser == null) {
      setWheelAxisHorizontal(true);
      setCursor(1, 1);
    } else {
      setWheelAxisHorizontal(parser.isWheelAxisHorizontal);
      if (parser.row != 0)
	setCursor(parser.row, parser.col);
      else
	setCursor(1, 1);
      if (parser.sheetName != null)
	this.sheetName = parser.sheetName;
      calculate();
    }
  }

  void calculateCellHeight() {
    showDescenders = ((flipsheet)Application.getCurrentApp()).showDescenders;
    cellHeight = valueFont.getAscent();
    if (showDescenders)
      cellHeight += valueFont.getDescent();
  }

  /**
   * Returns empty string if empty.
   */
  String toCSV() {
    StringBuffer sb = new StringBuffer();
    int lastCol = nCols;
    int lastRow = 0;
    // calculate lastRow.
    boolean cellSeen = false;
    {
      for (int row = maxRows-1; !cellSeen && (row > 0); row--) {
	for (int col = nCols-1; col > 0; col--) {
	  SheetTableCell td = (SheetTableCell)getTableCell(row, col);
	  String val = ((td == null) ? null : td.getContents());
	  if (val != null && val.length() > 0) {
	    cellSeen = true;
	    // don't change lastCol because we depend on having 
	    // a constant number of columns.
	    lastRow = row+1;
	    break;
	  } else {
	    // skip empty
	  }
	}
      }
    }
    
    // Structured comment
    // # !A1 Title
    // or
    // # -A1 Title
    {
      sb.append("# ");
      sb.append((isWheelAxisHorizontal()) ? "-" : "|");
      sb.append(getCellName());
      sb.append(" ");
      sb.append(getSheetName());
      sb.append("\n");
    }

    // Structured Comment
    // #.0 
    // to 
    // #.9
    // or more!
    {
      sb.append("#.");
      sb.append(getParser().getPrecision());
      sb.append("\n");
    }

    // Structured comments for onLoad actions.
    // #=expr2
    // #=expr1
    if (onLoad != null) {
      for (ListIterator it = onLoad.listIterator(); it.hasNext(); ) {
	String expr = (String)it.next();
	sb.append("#=");
	sb.append(expr);
	sb.append("\n");
      }
    }

    // if there's no cells and no onLoad, consider it empty and stop here.
    if (!cellSeen && onLoad == null) 
      return "";

    int lastCol1 = lastCol-1; 
    for (int i = 1; i < lastRow; i++) {
      for (int j = 1; j < lastCol; j++) {
	SheetTableCell td = (SheetTableCell)getTableCell(i, j);
	String contents = td.getContents();
	if (contents == null || contents.length() == 0) {
	  if (j != lastCol1)
	    sb.append(",");
	  else
	    sb.append("\n");
	} else {
	  sb.append("\"");
	  String s = td.getContents();
	  if (s.indexOf('"') == -1) {
	    sb.append(s);
	  } else {
	    int slen = s.length();
	    for (int si = 0; si<slen; si++) {
	      char c = s.charAt(si);
	      if (c == '"') sb.append('\\');
	      sb.append(c);
	    }
	  } 
	  if (j != lastCol1)
	    sb.append("\",");
	  else
	    sb.append("\"\n");
	}
      }
    }
    String v = sb.toString();
    DEBUG.p("v.length="+v.length());
    return v;
  }

  final String evaluate(String f) {
    String value;
    try {
      Object result = getParser().toplevelParse(f).eval();
      value = result.toString();
      if (value == null) DEBUG.p("null result for " + f);
    } catch (Exception e) {
      DEBUG.p("exception " + e);
      e.printStackTrace();
      value = e.getMessage();
    }
    return value;
  }

  final Pratt getParser() {
    if (parser == null)
      parser = new Pratt(this);
    return parser;
  }    

  final void setPrecision(int precision) {
    getParser().setPrecision(precision);
  }

  final int getPrecision() {
    return getParser().getPrecision();
  }

  LinkedList calculateDependencies(String f) {
    String value;
    try {
      return getParser().toplevelParse(f).getCellRefs();
    } catch (Exception e) {
      DEBUG.p("exception " + e);
      e.printStackTrace();
      return null;
    }
  }

//  public boolean eventWidgetUp(int i, Event event) {
//    DEBUG.p("SheetTable eventWidgetUp " + i);
//    event.debugPrint();
//    return super.eventWidgetUp(i, event);
//  }


}

// Except for empty values, we always require quote (") quoted values, and use ("") for quoting a quote.
// We might move to dispensing with quotes, especially on numbers and formulas.
final class CSVParser /* implements Iterator */ {
  final String csv;
  final int length;
  String sheetName;
  LinkedList onLoad;
  int row;
  int col;
  int precision = 9;
  boolean isWheelAxisHorizontal = true;

  int pos = 0;
  CSVParser (String csv) {
    this.csv = csv;
    this.length = csv.length();
    if (! parseComment()) {
      // if failed parsing comment try again as cell contents.
      pos = 0;
    }
  }

  // even if hasNext() is true, next might return "" anyway.
  public boolean hasNext() { 
    return pos < length; 
  }

  public Object next() {
    StringBuffer sb = new StringBuffer();
    boolean done = false;
    boolean inquote = false;
    while (pos < length && ! done) {
      char c = csv.charAt(pos);
      switch(c) {
      case '"':
	pos++;
	if (! inquote) {
	  inquote = true;
	} else {
	  if (pos >= length) {
	    done = true;
	  } else {
	    inquote = !inquote;
	  }
	}
	break;

      case '\\':
	// \" is quoted quote.
	pos++;
	if (inquote && pos<length) {
	  sb.append(csv.charAt(pos));
	  pos++;
	} 
	break;

	// both newline and comma terminate an entry outside strings.
      case '\n':
      case ',':
	if (inquote) {
	  sb.append(c);
	} else {
	  done = true;
	}
	pos++;
	break;

	// return char we insert in quoted string or ignore completely
      case '\r':
	if (inquote) {
	  sb.append(c);
	} 
	pos++;
	break;
      default:
	if (inquote) 
	  sb.append(c);
	else
	  DEBUG.p("flipsheet: Ignoring character " + c +" outside quotes");
	pos++;
      }
    }
    return sb.toString();
  }

  private boolean parseComment() {
    // If the first line doesn't start with a '#' then just give up now.
    if (pos>=length || csv.charAt(pos) != '#') 
      return false;
    // The right thing to would be to store the metadata into a different DataStore
    // and use UIDs to refer to the other entries,
    // and so the amount of data updated when you just move the cursor
    // and don't change the sheet is lower.  But we just store
    // them as comments in the serialized CSV.
    while (pos < length) {
      //DEBUG.p("Looking at char " + csv.charAt(pos)+" at position " + pos);
      if (csv.charAt(pos) != '#') return true;
      if (++pos >= length) return true;
      switch(csv.charAt(pos)) {
	// #<space>[|-]A42<space>My Spreadsheet Title<newline>
      case ' ': {
	if (++pos >= length) return false;
	// #<space>A42<space>My Spreadsheet Title<newline>
	//         P
	// #<space>[-|]A42<space>My Spreadsheet Title<newline>
	//         P
	char in = csv.charAt(pos);
	if (in == '-') {
	  isWheelAxisHorizontal = true;
	  pos++;
	} else if (in == '|') {
	  isWheelAxisHorizontal = false;
	  pos++;
	} 

	if (pos >= length) return false;
	in = Character.toUpperCase(csv.charAt(pos++));
	if ('A' <= in && in < (SheetTable.cellWidths.length+'@')) {
	  col = (in - 'A') + 1;
	} else {
	  return false;
	}

	// #<space>A42<space>My Spreadsheet Title<newline>
	//          P
	if (pos >= length) return false;
	int i = csv.indexOf(' ', pos);	  // skip over the space.
	if (i == -1) {
	  return false;
	}
	// #<space>A42<space>My Spreadsheet Title<newline>
	//          P I
	try {
	  if (pos >= length) return false;
	  row = Integer.parseInt(csv.substring(pos, i));
	  pos = i+1;
	} catch (NumberFormatException e) {
	  return false;
	}
	// #<space>A42<space>My Spreadsheet Title<newline>
	//            I      P
	StringBuffer sb = new StringBuffer();
	while(pos < length && (csv.charAt(pos) != '\r' && csv.charAt(pos) != '\n')) {
	  sb.append(csv.charAt(pos));
	  pos++;
	  if (pos >= length) return false;
	}
	skipEOL();
	sheetName = sb.toString();
	break;
      }

	// Look for optional lines starting with '#=', which are expressions to execute after load.
	// (Though they really ought to have dependencies).
      case '=': {
	pos++;
	int eol = csv.indexOf('\n', pos);
	if (eol == -1) eol = csv.length();
	if (onLoad == null) onLoad = new LinkedList();
	String expr = csv.substring(pos, eol).trim();
	if (expr.length() > 0) {
	  onLoad.add(expr);
	}
	pos = eol;
	skipEOL();
	break;
      }

	// #.precision
      case '.': {
	pos += 1;
	int eol = csv.indexOf('\n', pos);
	if (eol == -1) eol = csv.length();
	if (onLoad == null) onLoad = new LinkedList();
	String expr = csv.substring(pos, eol).trim();
	if (expr.length() > 0) {
	  precision = Integer.valueOf(expr).intValue();
	}
	pos = eol;
	skipEOL();
	// DEBUG.p("precision set to " + precision);
	break;
      }

      default: {
	DEBUG.p("did not understand char " + csv.charAt(pos) + " at column " + pos);
	skipEOL();
	break;
      }
      }
    }
    return true;
  }

  void skipEOL() {
    // eat through \r or \n left over
    while(pos < length && ((csv.charAt(pos) == '\r' || csv.charAt(pos) == '\n')))
      pos++;
  }
}
