package com.hipme.flipsheet;

import danger.app.Application;
import danger.util.LinkedList;
import danger.util.StringTokenizer;
import danger.util.DEBUG;

// Except for empty values, we always require quote (") quoted values, and use ("") for quoting a quote.
// We might move to dispensing with quotes, especially on numbers and formulas.
final class CSVParser /* implements Iterator */ {
  private static final boolean debugcsv = false;
  final String csv;
  final int length;
  String sheetName;
  LinkedList onLoad;
  int[] cellWidths;
  int row;
  int col;
  int precision = 9;
  int nRows;
  boolean isWheelAxisHorizontal = true;

  int pos = 0;
  CSVParser (String csv) {
    this.csv = csv;
    this.length = csv.length();
    if (! parseComment()) {
      if (debugcsv) DEBUG.p("parsing initial comment failed; trying from pos=0");
      // if failed parsing comment try again as cell contents.
      pos = 0;
    }
  }

  // even if hasNext() is true, next might return "" anyway.
  public boolean hasNext() { 
    return pos < length; 
  }

  public Object next() {
    StringBuffer sb = new StringBuffer();
    boolean done = false;
    boolean inquote = false;
    while (pos < length && ! done) {
      char c = csv.charAt(pos);
      switch(c) {
      case '"':
	pos++;
	if (! inquote) {
	  inquote = true;
	} else {
	  if (pos >= length) {
	    done = true;
	  } else {
	    inquote = !inquote;
	  }
	}
	break;

      case '\\':
	// \" is quoted quote.
	pos++;
	if (inquote && pos<length) {
	  sb.append(csv.charAt(pos));
	  pos++;
	} 
	break;

	// both newline and comma terminate an entry outside strings.
      case '\n':
      case ',':
	if (inquote) {
	  sb.append(c);
	} else {
	  done = true;
	}
	pos++;
	break;

	// return char we insert in quoted string or ignore completely
      case '\r':
	if (inquote) {
	  sb.append(c);
	} 
	pos++;
	break;
      default:
	if (inquote) 
	  sb.append(c);
	else
	  DEBUG.p("flipsheet: Ignoring character " + c +" outside quotes");
	pos++;
      }
    }
    return sb.toString();
  }

  private boolean parseComment() {
    // If the first line doesn't start with a '#' then just give up now.
    if (pos>=length || csv.charAt(pos) != '#') {
      if (debugcsv) DEBUG.p("flipsheet parseComment: 1; pos="+pos+" length="+length+" charAt(pos)="+csv.charAt(pos)+" csv="+csv);
      return false;
    }
    // The right thing to would be to store the metadata into a different DataStore
    // and use UIDs to refer to the other entries,
    // and so the amount of data updated when you just move the cursor
    // and don't change the sheet is lower.  But we just store
    // them as comments in the serialized CSV.
    while (pos < length) {
      // DEBUG.p("Looking at char " + csv.charAt(pos)+" at position " + pos + " length = " + length);
      // Skip CR or LF (sometimes we get both)
      if ((csv.charAt(pos) == '\r') || (csv.charAt(pos) == '\n')) {
	pos++;
	continue;
      }
      if (csv.charAt(pos) != '#') {
	if (debugcsv) DEBUG.p("flipsheet parseComment: 2 " +((int)csv.charAt(pos)));
	return true;
      }
      if (++pos >= length) {
	if (debugcsv) DEBUG.p("flipsheet parseComment: 3");
	return true;
      }
      switch(csv.charAt(pos)) {
	// #<space>[|-]A42<space>My Spreadsheet Title<newline>
      case ' ': {
	if (++pos >= length) {
	  if (debugcsv) DEBUG.p("flipsheet parseComment: 4");
	  return false;
	}
	// #<space>A42<space>My Spreadsheet Title<newline>
	//         P
	// #<space>[-|]A42<space>My Spreadsheet Title<newline>
	//         P
	char in = csv.charAt(pos);
	if (in == '-') {
	  isWheelAxisHorizontal = true;
	  pos++;
	} else if (in == '|') {
	  isWheelAxisHorizontal = false;
	  pos++;
	} 

	if (pos >= length) {
	  if (debugcsv) DEBUG.p("flipsheet parseComment: 5");
	  return false;
	}
	in = Character.toUpperCase(csv.charAt(pos++));
	if ('A' <= in && in <= (((char)SheetTable.MAX_COLS)+'@')) {
	  col = (in - 'A') + 1;
	} else {
	  if (debugcsv) DEBUG.p("flipsheet parseComment: 6: "+ in);
	  return false;
	}

	// #<space>A42<space>My Spreadsheet Title<newline>
	//          P
	if (pos >= length) {
	  if (debugcsv) DEBUG.p("flipsheet parseComment: 7");;
	  return false;
	}
	int i = csv.indexOf(' ', pos);	  // skip over the space.
	if (i == -1) {
	  if (debugcsv) DEBUG.p("flipsheet parseComment: 8");
	  return false;
	}
	// #<space>A42<space>My Spreadsheet Title<newline>
	//          P I
	try {
	  if (pos >= length) {
	    if (debugcsv) DEBUG.p("flipsheet parseComment: 9");
	    return false;
	  }
	  row = Integer.parseInt(csv.substring(pos, i));
	  pos = i+1;
	} catch (NumberFormatException e) {
	    if (debugcsv) DEBUG.p("flipsheet parseComment: 9.5");
	  if (debugcsv) e.printStackTrace();
	  return false;
	}
	// #<space>A42<space>My Spreadsheet Title<newline>
	//            I      P
	StringBuffer sb = new StringBuffer();
	while(pos < length && (csv.charAt(pos) != '\r' && csv.charAt(pos) != '\n')) {
	  sb.append(csv.charAt(pos));
	  pos++;
	  if (pos >= length) {
	    if (debugcsv) DEBUG.p("flipsheet parseComment: 10");
	    return false;
	  }
	}
	skipEOL();
	sheetName = sb.toString();
	break;
      }

     // Set number of rows and columns.
     // First number is number of rows.
     // Remaining numbers are pixel widths of columns.  First column is heading.
     // Ignore and use defaults if there's a problem.
      case '#': {
	pos++;
	int eol = findEOL();
	if (eol == -1) eol = csv.length();
	String data = csv.substring(pos, eol);
	if (debugcsv) DEBUG.p("##" + data);
	int[] cw = null;
	int nr;
	{
	  int i = 0;
	  StringTokenizer st = new StringTokenizer(data, " \r\n");
	  while (st.hasMoreTokens()) {
	    st.nextToken();
	    i++;
	  }
	  // Must have at least 1 row number 2 columns
	  int nc = i - 1;
	  if (nc >= 2)
	    cw = new int[nc];
	}
	{
	  int i = 0;
	  int w = 0;

	  StringTokenizer st = new StringTokenizer(data, " \r\n");
	  nr = Integer.valueOf(st.nextToken().trim()).intValue();
	  try {
	    while (st.hasMoreTokens()) {
	      int val = Integer.valueOf(st.nextToken().trim()).intValue();
	      w += val;
	      cw[i++] = val;
	    }
	  } catch (NumberFormatException fe) {
	    if (debugcsv) DEBUG.p("flipsheet: ## " + fe);
	    cw = null;
	  }
	  if ((w != SheetTable.TABLE_WIDTH) && ((flipsheet)Application.getCurrentApp()).exactTableWidth) {
	    SheetWindow.errorprint("table width " + w + "!="+SheetTable.TABLE_WIDTH+"; ignoring");
	    cw = null;
	  }
	  if (nr > SheetTable.MAX_ROWS) {
	    SheetWindow.errorprint("nRows="+nr+">"+SheetTable.MAX_ROWS+"; ignoring");
	    cw = null;
	  }
	}

	// If all went OK, set the width and rows.
	if (cw != null) {
	  cellWidths = cw;
	  nRows = nr;
	}
	skipEOL();
	break;
      }

     // Look for optional lines starting with '#=', which are expressions to execute after load.
     // (Though they really ought to have dependencies).
      case '=': {
	pos++;
	int eol = findEOL();
	if (onLoad == null) onLoad = new LinkedList();
	String expr = csv.substring(pos, eol).trim();
	if (expr.length() > 0) {
	  onLoad.add(expr);
	}
	skipEOL();
	break;
      }

	// #.precision
      case '.': {
	pos += 1;
	int eol = findEOL();
	if (onLoad == null) onLoad = new LinkedList();
	String p = csv.substring(pos, eol).trim();
	if (p.length() > 0) {
	  // DEBUG.p("precision = '"+p+"' pos="+pos+" eol="+eol);
	  precision = Integer.valueOf(p).intValue();
	}
	skipEOL();
	// DEBUG.p("precision set to " + precision);
	break;
      }

      default: {
	DEBUG.p("did not understand char " + csv.charAt(pos) + " at column " + pos);
	skipEOL();
	break;
      }
      }
    }
    return true;
  }

  // eat through \r or \n left over
  private void skipEOL() {
    pos = findEOL()+1;
    if (pos > length) pos = length;
  }

  private int findEOL() {
    int i = pos;
    while(i < length && ((csv.charAt(i) != '\r' && csv.charAt(i) != '\n'))) {
      i++;
    }
    return i;
  }

}
