package com.hipme.flipsheet;

import danger.app.*;
import danger.internal.Date;
import danger.ui.*;
import danger.ui.StaticText;
import danger.audio.Meta;
import danger.system.*;
import java.util.Vector;
import java.util.Enumeration;
import danger.util.DEBUG;
import danger.util.LinkedList;
import danger.util.ListIterator;
import com.hipme.flipsheet.parser.Pratt;
import com.hipme.flipsheet.parser.EvaluatorResult;

public final class SheetTableCell extends MyTableCell {
  String formula;
  int row;
  int col;
  LinkedList dependents;
  public int backgroundColor = Color.WHITE;
  public int foregroundColor = Color.BLACK;
  private boolean highlighted = false;
  int flags = 0;
  final static int CELL_LOCKED = 1;

  SheetTableCell(SheetTable sheetTable, int row, int col) {
    super(sheetTable);
    this.row = row;
    this.col = col;
    this.flags = 0;
  }

  String getFormula() {
    return formula;
  }

  String getContents() {
    if (formula != null) 
      return formula;
    else
      return getElementAsString();
  }

  void setCellContents(String newValue) {
    LinkedList oldDependencies = getDependencies();
    LinkedList newDependencies = getDependencies(newValue);
    if (! equalDependencies(oldDependencies, newDependencies)) {
      // If the dependencies aren't equal, update the SheetTableCells pointers to each other
      // and invalidate the master dependencies.  
      removeDependencies(oldDependencies);
      addDependencies(newDependencies);
      sheetTable.invalidateMasterDependencies();
    }
    if (newValue != null && (newValue.length()>0) && (newValue.charAt(0) == '=')) {
      // If the old formula was null and is now going to be non-null, even if there
      // are no dependencies on other cells, we have to add this cell to the
      // master dependencies list.  If the cell has any dependencies, it will
      // happen automatically, but if it has no dependencies,
      // The easiest thing to do is to invalidate the master dependencies
      // and let that take care of it.
      if (this.formula == null && (newDependencies == null || newDependencies.size()==0))
	sheetTable.invalidateMasterDependencies();
      this.formula = newValue;
      setElementToString("");
    } else {
      this.formula = null;
      setElementToString(newValue);
      if (isNumber(newValue)) {
	setAlignment(StaticText.HALIGN_RIGHT);
      } else {
	setAlignment(StaticText.HALIGN_LEFT);
      }
    }
  }

  LinkedList getDependencies() {
    return getDependencies(formula);
  }

  LinkedList getDependencies(String f) {
    if (f == null) return null;
    if (!f.startsWith("=")) return null;
    return sheetTable.calculateDependencies(f);
  }

  private void removeDependencies(LinkedList dependencies) {
    if (dependencies == null) return;
    for (ListIterator it = dependencies.listIterator(); it.hasNext(); ) {
      Pratt.CellRef c = ((Pratt.CellRef)it.next());
      SheetTableCell other = ((SheetTableCell)(sheetTable.getTableCell(c.getRow(), c.getColumn())));
      other.removeDependent(this);
    }
  }

  private void addDependencies(LinkedList dependencies) {
    if (dependencies == null) return;
    for (ListIterator it = dependencies.listIterator(); it.hasNext(); ) {
      Pratt.CellRef c = ((Pratt.CellRef)it.next());
      SheetTableCell other = ((SheetTableCell)(sheetTable.getTableCell(c.getRow(), c.getColumn())));
      if (other == null) DEBUG.p("!!! " +((char)(c.getRow()+'A')) + c.getColumn() + " was null");
      other.addDependent(this);
    }
  }

  private void addDependent(SheetTableCell cell) { 
    if (dependents == null) 
      dependents = new LinkedList(); 
    dependents.add(cell); 
  }

  private void removeDependent(SheetTableCell cell) { 
    if (dependents != null)
      dependents.remove(cell); 
  }

  LinkedList getDependents() { 
    return dependents; 
  }

  /**
   * Returns false if the two dependencies are not equal.
   * If they aren't in the same order, returns false as well -- too hard to optimize.
   */
  boolean equalDependencies(LinkedList a, LinkedList b) {
    if (a==null) return (b == null);
    if (b==null) return false;
    if (a.size() != b.size()) return false;
    ListIterator ai = a.listIterator();
    ListIterator bi = b.listIterator();
    while (ai.hasNext()) {
      if (ai.next() != bi.next())
	return false;
    }
    return true;
  }

  void calculate() {
    if (formula != null) {
      Object result = evaluateToObject();
      String newValue = result.toString();
      if (newValue == null) {
	newValue = "";
	DEBUG.p("formula " + formula + " evaluated to null");
      }
      if (! newValue.equals(getElementAsString()))
	setElementToString(newValue);
      if ((result instanceof EvaluatorResult) && ((EvaluatorResult)result).isNumber()) {
	setAlignment(StaticText.HALIGN_RIGHT);
      } else {
	setAlignment(StaticText.HALIGN_LEFT);
      }
    }
  }

  private Object evaluateToObject() {
    if (formula.length() == 0) return formula;
    if (formula.charAt(0) == '=') {
      return sheetTable.evaluateToObject(formula);
    }
    return formula;		// shouldn't happen
  }


  private boolean isNumber(String string) {
    int len = string.length();
    if ((len > 0) && Character.isDigit(string.charAt(0))) return true;
    if ((len > 1) && (string.charAt(0)=='.') && (Character.isDigit(string.charAt(1)))) return true;
    return false;
  }


  public String getName() {
    return ((char)('A' + col - 1)) + "" + row;
  }


  public String toString() { 
    return getName();
  }

  boolean isHighlighted() {
    return highlighted;
  }

  void setHighlighted(boolean t) {
    if (!sheetTable.useOutline) {
      if (t) {
	setBackgroundColor(SheetTable.highlightBackgroundColor);
      } else {
	setBackgroundColor(backgroundColor);
      }
    }
    highlighted = t;
  }

  boolean isLocked() {
    return ((flags & CELL_LOCKED)!=0);
  }

  void setLocked(boolean locked) {
    if (locked)
      flags = flags | CELL_LOCKED;
    else
      flags = flags & ~CELL_LOCKED;
  }

  public void paint(Pen pen, Rect rect) {
    super.paint(pen, rect);
    if (highlighted && sheetTable.useOutline)
      pen.drawRect(rect.left-1, rect.top-1, rect.left + getWidth()+2, rect.top + getHeight()+2);
  }

}

