package com.hipme.flipsheet;


import danger.app.*;
import danger.app.AppResources;
import danger.app.Application;
import danger.app.Event;
import danger.audio.Meta;
import danger.internal.Date;
import danger.system.*;
import danger.ui.*;
import danger.util.DEBUG;
import danger.util.format.StringFormat;
import java.util.Vector;

/**
 * Implements the main window class.
 *
 */
final class SheetWindow extends ScreenWindow implements Resources, Events, Constants {
  SheetScrollView sheetScrollView;
  SheetTable sheetTable;
  CellContentsTextField cellContentsTextField;
  String []sheetNamesCache = null;

  public SheetWindow(String csv) {
    buildWorld(csv);
    update();
  }
	
  void update(String csv) {		
    buildTable(csv);
    update();
  }

  void update() {		
    updateCellContents();
  }

  private void buildWorld(String csv) {
    setIcon(Application.getCurrentApp().getBitmap(AppResources.ID_SMALL_ICON));
    sheetScrollView = new SheetScrollView(getWidth(), getHeight());
    addChild(sheetScrollView);
    cellContentsTextField = new CellContentsTextField(this);
    cellContentsTextField.setEditable(false);
    addChild(cellContentsTextField);
    cellContentsTextField.show();
    buildTable(csv);
    sheetScrollView.show();
    if (getFocusedChild() == null) {
      setFocusedChild(sheetScrollView);
    }
  }

  private void buildTable(String csv) {
    sheetTable = new SheetTable(csv);
    sheetScrollView.setSheetTable(sheetTable);
    sheetTable.show();
  }

  public void adjustActionMenuState() {
    ResourceDatabase resource = Application.getCurrentApp().getResources();
    Menu actionMenu = getActionMenu();
    actionMenu.removeAllItems();
    if (getFocusedChild() == cellContentsTextField) 
      actionMenu.addFromResource(resource, kMainMenuCancelResID, this);
    MenuItem item = actionMenu.addItem(resource.getString(kStrFlipsheetsMenu));
    Menu subMenu = resource.getMenu(kFlipsheetsSubMenu);
    adjustSheetMenuItems(subMenu);
    item.addSubMenu(subMenu);
    actionMenu.addFromResource(resource, kMainMenuResID, this);
  }

  String getCSV() {
    return getCSV(0, 0, false, false);
  }

  String getCSV(int rowDelta, int colDelta, boolean insertOnLeft, boolean makeTableWider) {
    // Commit any outstanding edits before suspend, save, mail, etc.
    // crock -- but events are too late.
    if (cellContentsTextField != null && 
	cellContentsTextField == getFocusedChild()) {
      editCellContentsDone(cellContentsTextField.toString());
    }
    return sheetTable.toCSV(rowDelta, colDelta, insertOnLeft, makeTableWider);
  }

  /**
   * Call to insert or discard rows at current row.
   * @param rowDelta if positive, add new rows, with current row as first new blank row.
   * If negative, discard -rowDelta rows, including current row.
   * @param colDelta
   */
  void addDiscardRowsColumns(int rowDelta, int colDelta, boolean insertOnLeft, boolean makeTableWider) {
    update(getCSV(rowDelta, colDelta, insertOnLeft, makeTableWider));
  }

  public boolean eventShortcut(char c, Event e) {
    switch (c) {
      // We only want Menu-Shift-T here, not Menu-t.
    case 'T': 
      toggleFullScreen();
      return true;
      // Menu-Shift-D invalidates dependency list.
    case 'D': 
      sheetTable.invalidateMasterDependencies();
      return true;
    case 'n': // Menu-Alt-n is Menu-<
      if (e.modifierIsActive(Event.EVENT_MODIFIER_ALT)) {
	return doShiftColumnLeft(1);
      } 
      break;
    case 'm':// Menu-Alt-m is Menu->
      if (e.modifierIsActive(Event.EVENT_MODIFIER_ALT)) {
	return doShiftColumnRight(1);
      } 
      break;
    default:
      break;
    }
    return super.eventShortcut(c, e);
  }

  public boolean receiveEvent(Event e) {
    flipsheet app = getFlipsheet();
    boolean consumed = true;
    switch (e.type) {
    case kAboutEvent:
    case kWelcomeEvent:
    case kHelpEvent:
    case kDisplaySettingsEvent:
    case kShowCellSettingsEvent:
    case kConfirmDiscardSheetEvent:
    case kInsertRowsEvent:
    case kShowInsertColumnsDialogEvent:
    case kInsertColumnsEvent:
    case kConfirmDiscardRowsEvent:
    case kConfirmDiscardColumnsEvent:
    case kDiscardRowsEvent:
    case kDiscardColumnsEvent:
    case kDiscardSheetEvent:
    case kShowFlipOutFlipSheetEvent:
    case kEditOnLoad:
    case kImportFlipSheetEvent:
    case kShowSheetSettingsEvent:
    case kShowColumnSettingsEvent:
    case kFlipOutFlipSheetEvent:
    case kFlipInFlipSheetEvent:
    case kMailFlipSheetEvent:
    case kNewFlipSheet:
    case kSetFlipSheet:
      consumed = app.receiveEvent(e);
      break;
    case kEditCell:
      doEditCell(e);
      break;
    case kEditCellDone:
      doEditCellDone(e);
      break;
    case kEditCellCancel:
      cancelEdit();
      break;
    case kUpdateCell:
      updateCellContents();
      break;
    case kToggleTiling:
      doToggleTiling();
      break;
    case kColumnSettingsDialogShiftLeft:
      doShiftColumnLeft(e);
      break;
    case kColumnSettingsDialogShiftRight:
      doShiftColumnRight(e);
      break;
    case kColumnSettingsDialogFitToCell:
      doFitColumnToCell(e);
      break;
    default:
      consumed = super.receiveEvent(e);
    }
    return consumed;
  }

  private void processMotion(int what) {
    switch(what) {
    case WHAT_DOWN_OR_RIGHT_MOTION:
      if (sheetTable.isWheelAxisHorizontal()) {
	do {
	  sheetTable.forwardCell();
	} while(sheetTable.isCellLocked() && sheetTable.getRowCursor() < sheetTable.nRows-1);
      } else {
	do {
	  sheetTable.nextRow();
	} while (sheetTable.isCellLocked() && sheetTable.getRowCursor() < sheetTable.nRows-1);
      }
      break;
    case WHAT_DOWN_MOTION:
      do {
	sheetTable.nextRow();
      } while (sheetTable.isCellLocked() && sheetTable.getRowCursor() < sheetTable.nRows-1);
      break;
    case WHAT_UP_MOTION:
      do {
	sheetTable.previousRow();
      } while (sheetTable.isCellLocked() && sheetTable.getRowCursor() >= 1);
      break;
    }
  }

  // If I don't capture the keyDown instgead of keyUp
  // then when I press enter
  // the textfield gets the key-up event and
  // exits focus on it.
  public boolean eventKeyUp(char c, Event e) {
    boolean consumed = true;
    if (cellContentsTextField != null && (getFocusedChild() == cellContentsTextField)) {
      consumed = super.eventKeyUp(c, e);
    } else if (c == 13) {
      editCellContents(true);
    } else {
      consumed = super.eventKeyUp(c, e);
    }
    return consumed;
  }

  public boolean eventKeyDown(char c, Event e) {
    if (cellContentsTextField != null && (getFocusedChild() == cellContentsTextField))
      return super.eventKeyDown(c, e);
    boolean consumed = true;
    switch(c) {
    case ' ':
      sheetTable.setWheelAxisHorizontal(!sheetTable.isWheelAxisHorizontal());
      break;
    case '':
      editCellContents(false, "");		// should be done with a shortcut?
      break;
    default:
      consumed = false;
      break;
    }
    // Plain characters enter and start typing.
    if (!consumed && c > ' ' && c <= '~') {
      String s = sheetTable.getCellContents();
      if (s == null) s = "";
      // We have to send the whole event to the cell since modifier keys can be pressed and released.
      // Is there a better way to defer the event processing until after the textfield gets focus?
      editCellContents(false, s, c, e);
      consumed = true;
    }
    if (!consumed) {
      consumed = super.eventKeyUp(c, e);
    }
    return consumed;
  }

  private void doEditCell(Event e) {
    if (e.argument == null) {
      if (sheetTable.isCellLocked()) 
	processMotion(e.what);	  
      else
	editCellContents(e.what == WHAT_DOWN_OR_RIGHT_MOTION);
    } else {
      if (sheetTable.isCellLocked()) 
	Meta.play(Meta.BEEP_COMMAND_REJECTED);
      else
	editCellContents((e.what == WHAT_DOWN_OR_RIGHT_MOTION), ((String)e.argument));
    }
  }

  final void cancelEdit() {
    updateCellContents();
    setFocusedChild(sheetScrollView);
  }

  private void doToggleTiling() {
    if (isTiled()) {
      sheetScrollView.setTiled(false);
      updateCellContents();
    } else {
      cellContentsTextField.show();
      cellContentsTextField.setEditable(false);
      sheetScrollView.setTiled(true);
      updateCellContents();
    }
  }

  boolean isTiled() {
    return sheetScrollView.isTiled();
  }

  private int getSheetNumber() {
    return getFlipsheet().getSheetNumber();
  }

  void setSheetName(String name) {
    if (name == null) return;
    getSheetNamesCache()[getSheetNumber()] = name;
    sheetTable.setSheetName(name);
    setTitle(name);
  }

  void invalidateSheetNames() {
    sheetNamesCache = null;
  }

  String [] getSheetNamesCache() {
    flipsheet app = getFlipsheet();
    int count = app.getSheetCount();
    if (sheetNamesCache == null || sheetNamesCache.length != count) {
      sheetNamesCache = new String[count];
      for (int i = 0; i < count; i ++) {
	try {
	  DataStore ds = app.getDataStore();
	  sheetNamesCache[i]=new CSVParser(app.getCSV(i)).sheetName;
	} catch (Exception e) {
	  DEBUG.p("Ignoring exception " + e + " trying to get name for sheet " + i);
	  e.printStackTrace();
	}
	if (sheetNamesCache[i] == null)
	  sheetNamesCache[i] = sheetTable.getDefaultSheetName(i); // crock
      }
    }
    return sheetNamesCache;
  }

  private void adjustSheetMenuItems(Menu menu) {
    flipsheet app = getFlipsheet();
    String[] names = getSheetNamesCache();
    int count = sheetNamesCache.length;
    for (int i = 0; i < count; i++) {
      MenuItem m = menu.addItem(names[i],
				kSetFlipSheet,
				i,
				null,
				this);
      if (i < 10) m.setShortcut((char)('0'+i));
    }
    // put in "New Sheet" item if not at max.
    if (count < flipsheet.MAX_SHEETS) {
      menu.addItem(app.getString(kStrNewFlipsheet), 
		   kNewFlipSheet,
		   count,
		   null,
		   this).setShortcut('n');
    }
    MenuItem current = getSheetMenuItem(menu, getSheetNumber());
    if (current != null) current.setChecked(true);
  }

  private MenuItem getSheetMenuItem(Menu menu, int i) {
    if (i < 0 || i >= flipsheet.MAX_SHEETS) return null;
    return menu.getItemFromShortcut((char)('0' + i));
  }

  private void editCellContents(boolean selectAll) {
    if (sheetTable.isCellLocked()) {
      if (selectAll)
	processMotion(WHAT_DOWN_OR_RIGHT_MOTION);	  
      else
	Meta.play(Meta.BEEP_COMMAND_REJECTED);
    }
    else
      editCellContents(selectAll, sheetTable.getCellContents());
  }


  private void editCellContents(boolean selectAll, String value) {
    if (sheetTable.isCellLocked()) {
      Meta.play(Meta.BEEP_COMMAND_REJECTED);
      return;
    }
    cellContentsTextField.editCellContents(selectAll, value);
    setSubTitle(sheetTable.getCellName());
    invalidate();
  }

  /**
   * Event e is the read-ahead event that should go to the text field.
   */
  private void editCellContents(boolean selectAll, String value, char c, Event e) {
    if (sheetTable.isCellLocked()) {
      Meta.play(Meta.BEEP_COMMAND_REJECTED);
      return;
    }
    cellContentsTextField.editCellContents(selectAll, value);
    // crock to redirect the down and up events to the TextField now that we've captured the
    // up event.
    Event down = new Event(e);
    down.setListener(cellContentsTextField);
    down.type = Event.EVENT_KEY_DOWN;
    Event up = new Event(down);
    up.type = Event.EVENT_KEY_UP;
    cellContentsTextField.eventKeyDown(c, down);
    cellContentsTextField.eventKeyUp(c, up);
    setSubTitle(sheetTable.getCellName());
    invalidate();
  }

  private void updateCellContents() {
    String value = sheetTable.getCellContents();
    cellContentsTextField.setContents(value);
    if (! isTiled())
      cellContentsTextField.hide();
    if (sheetTable.isCellLocked()) {
      // so I'd like use use Unicode 29BB but it fails
      // so I'll use this X mark.  Crock!
      setSubTitle(sheetTable.getCellName() + " " + danger.ui.Font.GLYPH_BACK);
    } else {
      setSubTitle(sheetTable.getCellName());
    }
  }

  private void doEditCellDone(Event e) {
    editCellContentsDone((String)(e.argument));
    if (e.what == WHAT_NO_MOTION) {
      updateCellContents();
    } else {
      processMotion(e.what);
      updateCellContents();
      sheetScrollView.keepCellInView();
    }
  }

  private void editCellContentsDone(String value) {
    setFocusedChild(sheetScrollView); // this scrolls to sheetTable.getInterestingRect()
    String oldValue = sheetTable.getCellContents();
    cellContentsTextField.setEditable(false);
    if (! value.equals(oldValue)) {
      sheetTable.setCellContents(value);
      sheetTable.calculate();
    }
  }

  private void toggleFullScreen() {
    setFullScreen(!isFullScreen());
    sheetScrollView.resize(getWidth(), getHeight());
    cellContentsTextField.resize();
    invalidate();
  }

  /**
   * @returns true if successful, false if not
   */
  private boolean doShiftColumnLeft(Event e) {
    int pixels = ((NumberEditor)((DialogWindow)e.argument).getDescendantWithID(kColumnSettingsDialogPixels)).getValue();
    return doShiftColumnLeft(pixels);
  }

  /**
   * @returns true if successful, false if not
   * Narrows the current column 1 pixel to the left
   * unless it's the last colum, in which case it widens
   * it one to the left.
   */
  private boolean doShiftColumnLeft(int pixels) {
    if (sheetTable.getColCursor() == sheetTable.nCols-1)
      return adjustColumnWidth(pixels, true);
    else
      return adjustColumnWidth(-pixels, false);
  }

  /**
   * @returns true if successful, false if not
   */
  private boolean doShiftColumnRight(Event e) {
    int pixels = ((NumberEditor)((DialogWindow)e.argument).getDescendantWithID(kColumnSettingsDialogPixels)).getValue();
    return doShiftColumnRight(pixels);
  }
  /**
   * @returns true if successful, false if not
   * Widens the current column 1 pixel to the right
   * unless it's the last column, in which case it narrow
   * it one from the left.
   */
  private boolean doShiftColumnRight(int pixels) {
    if (sheetTable.getColCursor() == sheetTable.nCols-1)
      return adjustColumnWidth(-pixels, true);
    else
      return adjustColumnWidth(pixels, false);
  }

  /**
   * @returns true if successful, false if not
   */
  boolean adjustColumnWidth(int pixels, boolean left) {
    boolean ok = sheetTable.adjustColumnWidth(pixels, left);
    if (ok)
      update(getCSV());
    else
      Meta.play(Meta.BEEP_COMMAND_REJECTED);
    return ok;
  }

  /**
   * @returns true if successful, false if not.
   */
  boolean doFitColumnToCell(Event e) {
    boolean ok = sheetTable.adjustFitColumnToCell();
    if (ok)
      update(getCSV());
    else
      Meta.play(Meta.BEEP_COMMAND_REJECTED);
    return ok;
  }

  ///////////////
  /// Utilities
  ///////////////

  private static flipsheet getFlipsheet() {
    return ((flipsheet)Application.getCurrentApp());
  }

  static void errorprint(String message) {
    DEBUG.p("flipsheet error: " + message); // I18N
    AlertWindow a = getFlipsheet().getAlert(kErrorAlert, getFlipsheet());
    a.setMessage(message);
    a.show();
  }

  static void errorprint(Exception e) {
    errorprint(e.getMessage() == null ? 
	       StringFormat.fromRsrc(kStrInternalError, e.toString())
	       : e.getMessage());
    e.printStackTrace();
  }

  static void errorprint(Exception e, String message) {
    errorprint(e, message+ ": " + e.getMessage());
  }
}
