/*
 * A simple Pratt-Parser for SIOD: 2-FEB-90, George Carrette, GJC@PARADIGM.COM
 * Converted to Java by "Leigh L. Klotz, Jr" <klotz@graflex.org>
 *
 *                   COPYRIGHT (c) 1990 BY                       
 *     PARADIGM ASSOCIATES INCORPORATED, CAMBRIDGE, MASSACHUSETTS.
 *         See the source file SLIB.C for more information. 
 *
 *
 * Based on a theory of parsing presented in:                       
 *                                                                      
 *  Pratt, Vaughan R., ``Top Down Operator Precedence,''         
 *  ACM Symposium on Principles of Programming Languages         
 *  Boston, MA; October, 1973.                                   
 *                                                                      
 *
 * The following terms may be useful in deciphering this code:
 *
 * NUD -- NUll left Denotation (op has nothing to its left (prefix))
 * LED -- LEft Denotation      (op has something to left (postfix or infix))
 *
 * LBP -- Left Binding Power  (the stickiness to the left)
 * RBP -- Right Binding Power (the stickiness to the right)
 *
 */
package com.hipme.flipsheet.parser;

import danger.util.LinkedList;
import danger.util.StringUtils;
import java.util.Hashtable;
import com.ibm.icu.math.*;   // import BigDecimal and MathContext


import com.hipme.flipsheet.SheetTable;
import com.hipme.flipsheet.SheetTableCell;
import danger.ui.Color;		//  evil
import com.ibm.icu.math.MathContext;

public final class Pratt implements Constants {
  Tokenizer tokenizer;
  SheetTable env;
  public static final MathContext NINEDIGITS = 
    new MathContext(9, MathContext.PLAIN, false, MathContext.ROUND_HALF_UP);
  public static final MathContext UNLIMITED = 
    new MathContext(0, MathContext.PLAIN, false, MathContext.ROUND_HALF_UP);
  MathContext context = NINEDIGITS;

  public Pratt (SheetTable env) {
    this.env = env;
  }

  /**
   * Set the number of digits of precision for calculation.
   * 0 means unlimited, determined by the number of digits in the operands.
   */
  public void setPrecision(int digits) {
    if (digits == 0) context = UNLIMITED;
    else if (digits == 9) context = NINEDIGITS;
    else context = new MathContext(digits, MathContext.PLAIN, false, MathContext.ROUND_HALF_UP);
  }

  public int getPrecision() {
    return context.getDigits();
  }

  public boolean isNumber(String string) {
    return getTokenizer().isNumber(string);
  }

  public static abstract class Tree implements EvaluatorResult {
    Tree reverse() { return this; }
    public String toString() { 
      StringBuffer sb = new StringBuffer("");
      addToString(sb);
      return sb.toString();
    }

    abstract void addToString(StringBuffer sb);

    public Tree eval() { return this; }

    public final LinkedList getCellRefs() { 
      return addCellRefs(null);
    }
      
    LinkedList addCellRefs(LinkedList cellRefs) {
      return cellRefs;
    }

    public BigDecimal bigDecimalValue() { return BigDecimal.ZERO; }
    // public BigDecimal bigDecimalValue() { throw new RuntimeException("#NAN! " + this); }
    
    public boolean isNumber() { return false; }

  }

  public final static class Node extends Tree {
    Tree left;
    Node right;
    Node(Tree left, Node right) {
      this.left = left;
      this.right = right;
    }

    public Tree eval() {
      if (right == null) {
	// right null so evaluating left (parethesized expr)
	return left.eval();
      } else {
	// applying left to right
	if (left instanceof Primitive)
	  try {
	    return ((Primitive)left).apply(right);
	  } catch (ValueException v) {
	    System.err.println("valueexception="+v+" v.value="+v.value);
	    return ((Tree)(v.value));
	  }
	else {
	  throw new EvaluatorException("#FUNC! " + left.toString());
	}
      }
    }

    LinkedList addCellRefs(LinkedList cellRefs) {
      if (left != null) cellRefs = left.addCellRefs(cellRefs);
      if (right != null) cellRefs = right.addCellRefs(cellRefs);
      return cellRefs;
    }


    Tree reverse() {
      if (right == null) 
	return this;
      else
	return reverse1(this, null);
    }
    private static Tree reverse1(Node l, Node acc) {
      if (l == null) return acc;
      return reverse1((Node)(l.right), new Node(l.left, acc));
    }

    void addToString(StringBuffer sb) { 
      Node it = this;
      sb.append("(");
      while (true) {
	if (it.left == null) {
	  sb.append("nil ");
	} else {
	  it.left.addToString(sb);
	}
	if (it.right == null)
	  break;
	sb.append(" ");
	if (! (it.right instanceof Node)) {
	  sb.append(". ");
	  it.right.addToString(sb);
	  break;
	}
	it = (Node)(it.right);
      }
      sb.append(")");
    }
  }

  private static abstract class Token extends Tree implements Constants {
    int getLBP() { return DEFAULT_LBP; }
    int getRBP() { return DEFAULT_RBP; }
    int getNUD() { return NONE; }
    int getLED() { return NONE; }
    Token getHeader() { return this; }
    Token getComma() { return COMMA; }
    Token getMatch() { return this; }

    // crocks
    static Token COMMA;
    static Token RPAR;
    static Token EOD;
    static BigDecimalToken ZERO;
    static BigDecimalToken ONE;
    static BigDecimalToken TEN;

    abstract String getSymbol();
    void addToString(StringBuffer sb) {
      sb.append(getSymbol());
    }
  }

  private static class Symbol extends Token {
    String symbol;
    String getSymbol() {return symbol;}

    Symbol(String symbol) {
      super();
      this.symbol = symbol;
    }
  }

  private class RefError extends Symbol {
    RefError() {
      super("#REF!");
    }
    Tree apply(Node args) {
      throw new ValueException(this);
    }
    public BigDecimal bigDecimalValue() { throw new ValueException(this); }
  }

  Tree apply(Node args) {
    throw new EvaluatorException("#APPLY!");
  }

  // should the public version of this be done with an interface?
  public class CellRef extends Symbol {
    int column;
    int row;
    public boolean isRowAbsolute() { return false; }
    public boolean isColumnAbsolute() { return false; }

    CellRef(String symbol) {
      super(symbol);
      column = Character.toUpperCase(symbol.charAt(0))-'A'+1;
      row = Integer.parseInt(symbol.substring(1));
    }

    public int getRow() { return row; }
    public int getColumn() { return column; }

    public boolean equal(Object b) {
      if (b == null || ! (b instanceof CellRef)) return false;
      CellRef other = (CellRef)b;
      return (other.row == row &&
	      other.column == column);
    }

    void addToString(StringBuffer sb) {
      if (false) {
	// someday perhaps figure out relative refs
	// if we know the current cell number
	// and use this representation.
	sb.append("RC[+/-");
	sb.append(column);
	sb.append(",+/-");
	sb.append(row);
	sb.append("]");
      } else {
	sb.append(((char)(column+'@')));
	sb.append(row);
      }
    }

    public Tree eval() { 
      SheetTableCell td = (SheetTableCell)(env.getTableCell(row, column));
      String str = null;
      if (td != null) 
	str = td.getElementAsString(); // implicit evaluation order
      if (str == null)
	str = "";
      return tokenizer.intern(str);
    }

    Tree apply(Node args) {
      return eval();
    }


    LinkedList addCellRefs(LinkedList cellRefs) {
      if (cellRefs == null) cellRefs = new LinkedList();
      cellRefs.add(this);
      return cellRefs;
    }

    public void reposition(StringBuffer sb, int rowDelta, int atRow, int colDelta, int atCol) {
      boolean bad = false;
      if (rowDelta < 0) {
	// If the reference is inside a range of deleted rows, replace it with #REF!
	if (row >= atRow && row < atRow+(-rowDelta)) {
	  // System.err.println("#REF!: row="+row+" >= atRow="+atRow+" && row < atRow+(-rowDelta="+(-rowDelta)+")");
	  bad = true;
	} else {
	  // If the reference is before a range of deleted rows, leave it alone.
	  // If the reference is after a range of deleted rows, adjust by the delta.
	  sb.append((char)((column+'@')));
	  sb.append(row+((row>=atRow) ? rowDelta : 0));
	}
      } else if (rowDelta > 0) {
	// If the reference is before the start of inserted rows, leave it alone.
	// If the reference is after the start of inserted rows, adjust by the delta.
	sb.append((char)((column+'@')));
	sb.append(row+((row>=atRow) ? rowDelta : 0));
      } else if (colDelta < 0) {
	// If the reference is inside a range of deleted columns, replace it with #REF!
	if (column >= atCol && column < (atCol+(-colDelta))) {
	  // System.err.println("#REF!: column="+column+" >= atCol="+atCol+" && column < atCol+(-colDelta="+(-colDelta)+")");
	  bad = true;
	} else {
	  // If the reference is before a range of deleted columns, leave it alone.
	  // If the reference is after a range of deleted columns, adjust by the delta.
	  sb.append((char)((column+((column>=atCol) ? colDelta : 0)+'@')));
	  sb.append(row);
	}
      } else if (colDelta > 0) {
	// If the reference is before the start of inserted columns, leave it alone.
	// If the reference is after the start of of inserted columns, adjust by the delta.
	sb.append((char)((column+((column>=atCol) ? colDelta : 0)+'@')));
	sb.append(row);
      }
      if (bad)
	sb.append("#REF!");
    }
  }

  /**
   * For $A$1, A$1, and $A1.
   */
  public final class AbsoluteCellRef extends CellRef {
    private boolean rowIsAbsolute;
    private boolean columnIsAbsolute;
    public final boolean isRowAbsolute() { return rowIsAbsolute; }
    public final boolean isColumnAbsolute() { return columnIsAbsolute; }

    AbsoluteCellRef(String symbol) {
      super(StringUtils.stripChar(symbol, '$'));
      if (symbol.charAt(0)=='$') columnIsAbsolute = true;
      if (symbol.lastIndexOf('$') != 0) rowIsAbsolute = true;
    }


    void addToString(StringBuffer sb) {
      if (columnIsAbsolute) 
	sb.append('$');
      sb.append(((char)(column+'@')));
      if (rowIsAbsolute) 
	sb.append('$');
      sb.append(row);
    }

    public void reposition(StringBuffer sb, int rowDelta, int atRow, int colDelta, int atCol) {
      boolean bad = false;
      if (rowDelta < 0) {
	// If the reference is inside a range of deleted rows, replace it with #REF!
	if (row >= atRow && row < atRow+(-rowDelta)) {
	  // System.err.println("#REF!: row="+row+" >= atRow="+atRow+" && row < atRow+(-rowDelta="+(-rowDelta)+")");
	  bad = true;
	} else {
	  // If the reference is before a range of deleted rows, leave it alone.
	  // If the reference is after a range of deleted rows, adjust by the delta.
	  if (columnIsAbsolute) sb.append('$');
	  sb.append((char)((column+'@')));
	  if (rowIsAbsolute) sb.append('$');
	  sb.append(row+((!rowIsAbsolute && (row>=atRow)) ? rowDelta : 0));
	}
      } else if (rowDelta > 0) {
	// If the reference is before the start of inserted rows, leave it alone.
	// If the reference is after the start of inserted rows, adjust by the delta.
	if (columnIsAbsolute) sb.append('$');
	sb.append((char)((column+'@')));
	if (rowIsAbsolute) sb.append('$');
	sb.append(row+((!rowIsAbsolute && (row>=atRow)) ? rowDelta : 0));
      } else if (colDelta < 0) {
	// If the reference is inside a range of deleted columns, replace it with #REF!
	if (column >= atCol && column < (atCol+(-colDelta))) {
	  // System.err.println("#REF!: column="+column+" >= atCol="+atCol+" && column < atCol+(-colDelta="+(-colDelta)+")");
	  bad = true;
	} else {
	  // If the reference is before a range of deleted columns, leave it alone.
	  // If the reference is after a range of deleted columns, adjust by the delta.
	  if (columnIsAbsolute) sb.append('$');
	  sb.append((char)((column+((!columnIsAbsolute && (column>=atCol)) ? colDelta : 0)+'@')));
	  if (rowIsAbsolute) sb.append('$');
	  sb.append(row);
	}
      } else if (colDelta > 0) {
	// If the reference is before the start of inserted columns, leave it alone.
	// If the reference is after the start of of inserted columns, adjust by the delta.
	if (columnIsAbsolute) sb.append('$');
	sb.append((char)((column+((!columnIsAbsolute && (column>=atCol)) ? colDelta : 0)+'@')));
	if (rowIsAbsolute) sb.append('$');
	sb.append(row);
      }
      if (bad)
	sb.append("#REF!");
    }
  }

  private static abstract class Primitive extends Symbol {
    Primitive(String symbol, int lbp, int rbp, int nud, int led) {
      super(symbol);
      this.lbp=lbp;
      this.rbp=rbp;
      this.nud=nud;
      this.led=led;
    }
    String symbol;
    int lbp;
    int rbp;
    int nud;
    int led;
    Token header;
    Token comma;
    Token match;

    int getLBP() { return lbp; }
    int getRBP() { return rbp; }
    int getNUD() { return nud; }
    int getLED() { return led; }

    Token getComma() { 
      if (comma == null) 
	return comma;
      else
	return super.getComma(); 
    }

    Token getMatch() { 
      if (match != null)
	return match;
      else 
	return super.getMatch();
	   
    }
    Token getHeader() { 
      if (header != null) 
	return header;
      else
	return super.getHeader();
    }

    abstract Tree apply(Node args);
  }

  private static abstract class PrefixPrimitive extends Primitive {
    PrefixPrimitive(String symbol) {
      super(symbol, DEFAULT_LBP, DEFAULT_RBP, PARSE_PREFIX, NONE);
    }
  }


  private static class EOD extends Primitive {
    EOD() {
      super("", -1, DEFAULT_RBP, NONE, PRETERM_ERR);
    }
    Tree apply(Node args) { throw new ParserException("#EOD!"); }
  }

  private static class LPar extends Primitive {
    LPar() {
      super("(", DEFAULT_LBP, DEFAULT_RBP, OPEN_PAREN_NUD, OPEN_PAREN_LED);
    }

    Tree apply(Node args) { return ((Node)(args.right.left)).eval(); }
  }

  private static class RPar extends Primitive {
    RPar() {
      super(")", 5, DEFAULT_RBP, DELIM_ERR, ERB_ERR);
    }
    Tree apply(Node args) { throw new ParserException("#RPAR!"); }
  }

  private static class Comma extends Primitive {
    Comma() {
      super(",", 10, DEFAULT_RBP, NONE, DELIM_ERR);
    }
    Tree apply(Node args) { throw new ParserException("#COMMA!"); }
  }

  private static class Colon extends Primitive {
    Colon() {
      super(":", DEFAULT_LBP, DEFAULT_RBP, NONE, PARSE_INFIX);
    }
    Tree apply(Node args) { 
      // System.err.println("args.left="+args.left+ " " + args.left.getClass());
      // System.err.println("args.right="+args.right+" " + args.right.getClass());
      // System.err.println("args.right.left="+args.right.left + " " + args.right.left.getClass());
      CellRef start = (CellRef)(args.left);
      CellRef end = (CellRef)(args.right.left);
      return args;
    }
  }

  private class Minus extends Primitive {
    Minus() {
      super("-", 100, 100, PARSE_PREFIX, PARSE_NARY);
    }

    Tree apply(Node args) {
      BigDecimal val = BigDecimal.ZERO;
      if (args != null) {
	BigDecimal arg1 = args.left.eval().bigDecimalValue();
	args = ((Node)args.right);
	if (args == null) {
	  val = val.subtract(arg1, context);
	} else {
	  val = arg1;
	  while (args != null) {
	    val = val.subtract(args.left.eval().bigDecimalValue(), context);
	    args = ((Node)args.right);
	  }
	}
      }
      return new BigDecimalToken(val);
    }
  }

  private class Plus extends Primitive {
    Plus () {
      super("+", 100, 100, PARSE_PREFIX, PARSE_NARY);
    }
    Tree apply(Node args) {
      BigDecimal val = BigDecimal.ZERO;
      while (args != null) {
	val = val.add(args.left.eval().bigDecimalValue(), context);
	args = ((Node)args.right);
      }
      return new BigDecimalToken(val);
    }
  } 

  private class Times extends Primitive {
    Times() {
      super("*", 120, DEFAULT_RBP, NONE, PARSE_NARY);
    }
    Tree apply(Node args) {
      BigDecimal val = BigDecimal.ONE;
      while (args != null) {
	val = val.multiply(args.left.eval().bigDecimalValue(), context);
	args = ((Node)args.right);
      }
      return new BigDecimalToken(val);
    }
  }

  private class Equals extends MathComparatorPrimitive {
    Equals() {
      super("=", 80, 80, NONE, PARSE_INFIX);
    }
    protected boolean compare(BigDecimal arg1, BigDecimal arg2) {
      return (arg1.compareTo(arg2) == 0);
    }
  }

  private static class Exponent extends Primitive {
    Exponent() {
      super("^", 140, 139, NONE, PARSE_INFIX);
    }
    Tree apply(Node args) {
      BigDecimal arg1 = args.left.eval().bigDecimalValue();
      BigDecimal arg2 = args.right.left.eval().bigDecimalValue();
      return new BigDecimalToken(arg1.pow(arg2));
    }
  }

  private class Divide extends Primitive {
    Divide() {
      super("/", 120, 120, NONE, PARSE_INFIX);
    }
    Tree apply(Node args) {
      BigDecimal val = BigDecimal.ONE;
      if (args != null) {
	val = args.left.eval().bigDecimalValue();
	args = ((Node)args.right);
	while (args != null) {
	  val = val.divide(args.left.eval().bigDecimalValue(), context);
	  args = ((Node)args.right);
	}
      }
      return new BigDecimalToken(val);
    }
  }

  private static class Percent extends Primitive {
    Percent() {
      super("%", 120, 120, NONE, PARSE_POSTFIX);
    }
    Tree apply(Node arg) {
      BigDecimal val = arg.eval().bigDecimalValue();
      val = val.divide(BigDecimal.TEN).divide(BigDecimal.TEN, NINEDIGITS);
      return new BigDecimalToken(val);
    }
  }

  private abstract class MathComparatorPrimitive extends Primitive {
    MathComparatorPrimitive(String symbol, int lbp, int rbp, int nud, int led) {
      super(symbol, lbp, rbp, nud, led);
    }
    Tree apply(Node args) {
      BigDecimal arg1 = args.left.eval().bigDecimalValue();
      BigDecimal arg2 = args.right.left.eval().bigDecimalValue();
      if (compare(arg1, arg2))
	return Token.ONE;
      else
	return Token.ZERO;
    }
    abstract protected boolean compare(BigDecimal arg1, BigDecimal arg2);
  }

  private class Less extends MathComparatorPrimitive {
    Less() {
      super("<", 80, 80, NONE, PARSE_INFIX);
    }
    protected boolean compare(BigDecimal arg1, BigDecimal arg2) {
      return (arg1.compareTo(arg2) == -1);
    }
  }

  private class Greater extends MathComparatorPrimitive {
    Greater() {
      super(">", 80, 80, NONE, PARSE_INFIX);
    }
    protected boolean compare(BigDecimal arg1, BigDecimal arg2) {
      return (arg1.compareTo(arg2) == 1);
    }
  }

  private class GreaterEqual extends MathComparatorPrimitive {
    GreaterEqual() {
      super(">=", 80, 80, NONE, PARSE_INFIX);
    }
    protected boolean compare(BigDecimal arg1, BigDecimal arg2) {
      return (arg1.compareTo(arg2) >= 0);
    }
  }

  private class NotEqual extends MathComparatorPrimitive {
    NotEqual() {
      super("!=", 80, 80, NONE, PARSE_INFIX);
    }
    protected boolean compare(BigDecimal arg1, BigDecimal arg2) {
      return (arg1.compareTo(arg2) == 0);
    }
  }

  private class LessEqual extends MathComparatorPrimitive {
    LessEqual() {
      super("<=", 80, 80, NONE, PARSE_INFIX);
    }
    protected boolean compare(BigDecimal arg1, BigDecimal arg2) {
      return (arg1.compareTo(arg2) <= 0);
    }
  }

  private class Not extends Primitive {
    Not() {
      super("!", 70, 70, NONE, PARSE_PREFIX);
    }
    Tree apply(Node args) {
      BigDecimal val = (args.left.eval().bigDecimalValue());
      if (val.equals(BigDecimal.ZERO)) 
	return Token.ONE;
      else 
	return Token.ZERO;
    }
  }

  private class And extends Primitive {
    And() {
      super("&", 65, DEFAULT_RBP, NONE, PARSE_INFIX);
    }
    Tree apply(Node args) {
      while (args != null) {
	Tree val = args.left.eval();
	if (val.bigDecimalValue().equals(BigDecimal.ZERO))
	  return val;
	args = ((Node)args.right);
      }
      return Token.ONE;
    }
  }

  private class Or extends Primitive {
    Or() {
      super("|", 60, DEFAULT_RBP, NONE, PARSE_NARY);
    }
    Tree apply(Node args) {
      while (args != null) {
	Tree val = args.left.eval();
	if (! (val.bigDecimalValue().equals(BigDecimal.ZERO)))
	  return val;
	args = ((Node)args.right);
      }
      return Token.ONE;
    }
  }

  private static class If extends PrefixPrimitive {
    If() { super("if"); }
    Tree apply(Node args) {
      args = (Node)(args.left);		// why?
      BigDecimal pred = args.left.eval().bigDecimalValue();
      if (pred.compareTo(BigDecimal.ZERO) != 0) {
	return args.right.left.eval();
      } else {
	return args.right.right.left.eval();
      }
    }
  }

  private abstract class RangePrimitive extends PrefixPrimitive {
    RangePrimitive(String name) { super(name); }
    BigDecimal fetch(int row, int column) throws NumberFormatException {
      SheetTableCell td = (SheetTableCell)(env.getTableCell(row, column));
      String str = null;
      if (td != null) 
	str = td.getElementAsString(); // implicit evaluation order
      if (str == null || str.length() == 0)
	return BigDecimal.ZERO;
      else
	return new BigDecimal(str);
    }
  }

  private abstract class MathRangePrimitive extends RangePrimitive {
    MathRangePrimitive(String name) {
      super(name);
    }
    abstract BigDecimal initialValue();
    abstract BigDecimal operate(BigDecimal value1, BigDecimal value2);
    Tree apply(Node args) {
      BigDecimal val = initialValue();
      args = (Node)(args.left);
      while (args != null) {
	Tree arg1v = (Tree)args.left.eval();
	if (arg1v instanceof BigDecimalToken) {
	  val = operate(val, (((BigDecimalToken)arg1v).bigDecimalValue()));
	} else if (arg1v instanceof Node && ((Node)arg1v).left instanceof CellRef) {
	  CellRef start = (CellRef)(((Node)arg1v).left);
	  CellRef end = (CellRef)(((Node)arg1v).right.left);
	  if (start.column == end.column) {
	    for (int row = start.row; row <= end.row; row++) {
	      val = operate(val, fetch(row, start.column));
	    }
	  } else if (start.row == end.row) {
	    for (int column = start.column; column <= end.column; column++) {
	      val = operate(val, fetch(start.row, column));
	    }
	  } else {
	    for (int row = start.row; row <= end.row; row++) {
	      for (int column = start.column; column < end.column; column++) {
		val = operate(val, fetch(row, column));
	      }
	    }
	  } 
	} else {
	  System.err.println("skipping mathrange arg " + arg1v);
	}
	args = ((Node)(args.right));
      }
      return new BigDecimalToken(val);
    }
  }

  private class Sum extends MathRangePrimitive {
    Sum() { super("sum"); }
    final BigDecimal initialValue() { return BigDecimal.ZERO; }
    final BigDecimal operate(BigDecimal val1, BigDecimal val2) { return val1.add(val2, context); }
  }

  private class Product extends MathRangePrimitive {
    Product() { super("product"); }
    final BigDecimal initialValue() { return BigDecimal.ONE; }
    final BigDecimal operate(BigDecimal val1, BigDecimal val2) { return val1.multiply(val2, context); }
  }

  private class Min extends MathRangePrimitive {
    Min() { super("min"); }
    final BigDecimal initialValue() { return null; }
    final BigDecimal operate(BigDecimal val1, BigDecimal val2) { 
      if (val1 == null) return val2;
      else if (val1.compareTo(val2, context) == -1) return val1;
      else return val2;
    }
  }

  private class Max extends MathRangePrimitive {
    Max() { super("max"); }
    final BigDecimal initialValue() { return null; }
    final BigDecimal operate(BigDecimal val1, BigDecimal val2) { 
      if (val1 == null) return val2;
      else if (val1.compareTo(val2, context) == 1) return val1;
      else return val2;
    }
  }

  // could be done more quickly but with additional code
  private class Count extends MathRangePrimitive {
    Count() { super("count"); }
    final BigDecimal initialValue() { return BigDecimal.ZERO; }
    final BigDecimal operate(BigDecimal val1, BigDecimal val2) { 
      return val1.add(BigDecimal.ONE, UNLIMITED);
    }
  }

  // Still to come:
  // average, npv, lookup, na, error, pi, abs, int, exp, sqrt
  // ln, log10, sin, asin, cos, acos, tan, atan

  private static class Debug extends PrefixPrimitive {
    Debug() { super("debug"); }
    Tree apply(Node argv) {
      StringBuffer sb = new StringBuffer();
      Node args = (Node)(argv.left);		// why?
      sb.append("flipsheet debug: ");
      while (args != null) {
	Tree v1 = args.left.eval();
	args = args.right;
	sb.append(v1.toString());
	if (args != null) sb.append(", ");
      }
      System.err.println(sb.toString());
      return argv;		// gotta return something.
    }
  }

  private abstract class RangeIntOp extends PrefixPrimitive {
    RangeIntOp(String symbol) { super(symbol); }
    abstract void setRangeInt(int row, int col, int value);
    void processRange(int value, Node args) {
      // subclass should deal with first args and invoke super.apply(restargs)
      while (args != null) {
	Tree arg = args.left;
	if (arg instanceof CellRef) {
	  CellRef cell = (CellRef)arg;
	  setRangeInt(cell.getRow(), cell.getColumn(), value);
	} else {
	  Node arg1v = (Node)arg.eval();
	  CellRef start = (CellRef)(arg1v.left);
	  CellRef end = (CellRef)(arg1v.right.left);
	  // System.err.println("start = " + start +" end = " + end);
	  if (start.column == end.column) {
	    for (int row = start.row; row <= end.row; row++) {
	      setRangeInt(row, start.column, value);
	    }
	  } else if (start.row == end.row) {
	    for (int column = start.column; column <= end.column; column++) {
	      setRangeInt(start.row, column, value);
	    }
	  } else {
	    for (int row = start.row; row <= end.row; row++) {
	      for (int column = start.column; column <= end.column; column++) {
		setRangeInt(row, column, value);
	      }
	    }
	  }
	}
	args = args.right;
      }
    }
  }

  private abstract class SetColor extends RangeIntOp {
    SetColor(String symbol) { super(symbol); }
    final Tree apply(Node argv) {
      StringBuffer sb = new StringBuffer();
      Node args = (Node)(argv.left);		// why?
      Tree nColor = args.left.eval();
      int color = 0;
      if (nColor instanceof BigDecimalToken) {
	color = nColor.bigDecimalValue().intValue();
      } else if (nColor instanceof StringToken) {
	color = Color.fromString(((StringToken)nColor).stringVal());
      } else if (nColor instanceof Symbol) {
	color = Color.fromString(((Symbol)nColor).getSymbol());
      } else {
	throw new EvaluatorException("unknown color arg: " + nColor + " (" + nColor.getClass().getName() + ")");
      }
      processRange(color, args.right);
      return nColor;		// gotta return something
    }
  }

  private class SetBackgroundColor extends SetColor {
    SetBackgroundColor() { super("setbg"); }
    void setRangeInt(int row, int col, int color) {
      env.setCellBackgroundColor(row, col, color);
    }
  }

  private class SetForegroundColor extends SetColor {
    SetForegroundColor() { super("setcolor"); }
    void setRangeInt(int row, int col, int color) {
      env.setCellForegroundColor(row, col, color);
    }
  }


  private class LockRange extends RangeIntOp {
    LockRange() { super("lock"); }
    void setRangeInt(int row, int col, int value) {
      env.setLocked(row, col, value == 0 ? false : true);
    }
    final Tree apply(Node argv) {
      StringBuffer sb = new StringBuffer();
      Node args = (Node)(argv.left);		// why?
      Tree nValue = args.left.eval();
      int value = 0;
      if (nValue instanceof BigDecimalToken) {
	value = nValue.bigDecimalValue().intValue();
      } else {
	throw new EvaluatorException("unknown locked/unlocked arg" + nValue);
      }
      processRange(value, args.right);
      return nValue;		// gotta return something
    }
  }

  // Returned by CellRef and used by double-quote.
  private static class StringToken extends Symbol {
    StringToken(String string) {
      super(string);
    }
    String stringVal() { return getSymbol(); }
  }

  private static abstract class NumberToken extends Token {
  }

  private static final class IntegerToken extends NumberToken {
    int value;

    IntegerToken(String string) {
      super();
      value = Integer.parseInt(string);
    }

    public String toString() { 
      return String.valueOf(value);
    }

    String getSymbol() {
      return String.valueOf(value);
    }

    public int intVal() { return value; }
  }

  private static final class BigDecimalToken extends NumberToken {
    BigDecimal value;

    BigDecimalToken(BigDecimal value) {
      super();
      this.value = value;
    }

    BigDecimalToken(String string) {
      super();
      value = new BigDecimal(string);
    }

    public String toString() { 
      return value.toString();
    }

    String getSymbol() {
      return value.toString();
    }

    public BigDecimal bigDecimalValue() { return value; }

    public boolean isNumber() { return true; }
  }

  // BUG: Right now either handles rowDelta or colDelta but not both.
  public final String reposition(String string, int rowDelta, int atRow, int colDelta, int atCol) {
    if (tokenizer == null) tokenizer = new Tokenizer();
    tokenizer.setString(string);
    StringBuffer sb = new StringBuffer("=");
    Token t;
    if (rowDelta != 0 && colDelta != 0) 
      throw new RuntimeException("Cannot handle rowDelta = " +rowDelta+" colDelta="+colDelta);
    while ((t = tokenizer.readToken()) != null) {
      if (t instanceof CellRef) {
	CellRef cr=((CellRef)t);
	cr.reposition(sb, rowDelta, atRow, colDelta, atCol);
      } else {
	sb.append(t);
      }
    }

    // System.err.println("sb="+sb);
    return sb.toString();   
  }

  Tokenizer getTokenizer() {
    if (tokenizer == null) 
      tokenizer = new Tokenizer();
    return tokenizer;
  }

  public final Tree toplevelParse(String string) {
    getTokenizer().setString(string);
    return toplevelParse(tokenizer);
  }

  public final Tree toplevelParse(Tokenizer stream) {
    if (stream.peekToken() == null)
      return null;
    return parse(-1, stream);
  }

  private Tree parse(int rbpLevel, Tokenizer stream) {
    return parseLoop(rbpLevel, stream, nudcall(stream.readToken(), stream));
  }

  private int lbp(Token t) {
    if (t == null) return DEFAULT_LBP;
    return t.getLBP();
  }

  private int rbp(Token t) {
    if (t == null) return DEFAULT_RBP;
    return t.getRBP();
  }

  private int led(Token t) {
    if (t == null) return NONE;
    return t.getLED();
  }

  private int nud(Token t) {
    if (t == null) return NONE;
    return t.getNUD();
  }

  private Tree parseLoop(int rbpLevel, Tokenizer stream, Tree translation) {
    if (rbpLevel < lbp(stream.peekToken()))
      return parseLoop(rbpLevel, stream, 
		       ledcall(stream.readToken(), translation, stream));
    return translation;
  }

  private Tree nudcall(Token token, Tokenizer stream) {
    if (token instanceof NumberToken) 
      return token;
    int nud = nud(token);
    switch(nud) {
    default:
      throw new ParserException("#ERR! " + nud + " (not a prefix operator)");

    case OPEN_PAREN_NUD:
      return openParenNud(token, stream);

    case PARSE_PREFIX:
      return parsePrefix(token, stream);

    case DELIM_ERR:
      throw new ParserException("#ILLEGAL! " + token + " (use of delimiter)");
      
    case ERB_ERR:
      throw new ParserException("#MANY!" + token);

    case PRETERM_ERR:
      throw new ParserException("#EOD!" + token);

    case NONE: 
      int led = led(token);
      if (led != NONE) {
	throw new ParserException("#SYNTAX!" + token + " (not a prefix operator)");
      }
      return token;
    }
  }

  private Tree ledcall(Token token, Tree left, Tokenizer stream) {
    // This test should be !Symbol rather than =Number, 
    // but it's OK for now.
    if (token instanceof NumberToken) {
      throw new ParserException("#SYNTAX!" + token);
    }
    int led = led(token);
    switch(led) {
    case NONE:
      System.err.println("ledcall NONE");
      throw new ParserException("#SYNTAX!" + token);
    default:
      throw new IllegalArgumentException("#ERR!" + led);

    case OPEN_PAREN_LED:
      return openParenLed(token, left, stream);

    case PARSE_INFIX:
      return parseInfix(token, left, stream);

    case PARSE_POSTFIX:
      return parsePostfix(token, left);

    case PARSE_NARY:
      return parseNary(token, left, stream);

    case DELIM_ERR:
      System.err.println("ledcall DELIM_ERR");
      throw new ParserException("#SYNTAX!" + token);
    case ERB_ERR:
      throw new ParserException("#MANY!" + token);
    case PRETERM_ERR:
      System.err.println("ledcall PRETERM_ERR");
      throw new ParserException("#SYNTAX!" + token);
    }
  }

  private Tree parsePrefix(Token token, Tokenizer stream) {
    return new Node(token.getHeader(),
		    new Node(parse(rbp(token), stream), null));
  }

  private Node parseInfix(Token token, Tree left, Tokenizer stream) {
    return new Node(token.getHeader(),
		    new Node(left,
			     new Node(parse(rbp(token), stream), null)));
  }

  private Node parsePostfix(Token token, Tree left) {
    return new Node(token.getHeader(), new Node(left, null));
  }

  private Node parseNary(Token token, Tree left, Tokenizer stream) {
    return new Node(token.getHeader(), new Node(left, prsnary(token, stream)));
  }

  private Node prsnary(Token token, Tokenizer stream) {
    return prsnaryLoop(token, 
		       stream,
		       new Node(parse(rbp(token), stream), null));
  }

  private Node prsnaryLoop (Token token, Tokenizer stream, Node l) {
    if (token == (stream.peekToken())) {
      stream.readToken();
      return prsnaryLoop(token, stream, new Node(parse(rbp(token), stream), l));
    } else {
      return (Node)(l.reverse());
    }
  }


  //
  //  Tree parseMatchfix(Token token, Tree left, Tokenizer stream) {
  //    return new Node(token.getHeader(),
  //		    prsmatch(token.getMatch(),
  //			     token.getComma(),
  //			     stream));
  //  }
  //
  //  Tree parseMatchfixModified (Token token, Tokenizer stream) {
  //    return new Node(token.getHeader(),
  //		    prsmatchModified(token.getMatch(),
  //				     token.getComma(),
  //				     stream));
  //  }

  private Node prsmatch(Token token, Token comma, Tokenizer stream) {
    if (token == (stream.peekToken())) {
      stream.readToken();
      return null;
    } else {
      return prsmatchLoop (token, comma, stream,
			   new Node(parse(10, stream), null));
    }
  }  

  private Node prsmatchLoop (Token token, Token comma, Tokenizer stream, Node l) {
    if (token == (stream.peekToken())) {
      stream.readToken();
      return (Node)(l.reverse());
    } else if (comma == (stream.peekToken())) {
      stream.readToken();
      return prsmatchLoop(token, comma, stream,
			  new Node(parse(10, stream), l));
    } else {     
      throw new ParserException("#COMMA! "+ (stream.readToken()));
    }
  }

//  private Node prsmatchModified (Token token, Token comma, Tokenizer stream) {
//    if (token == (stream.peekToken())) {
//      stream.readToken();
//      return null;
//    } else {
//      return prsmatchModifiedLoop(token, comma, stream,
//				  new Node(parse(10, stream), null));
//    }
//  }
//
//  private Node prsmatchModifiedLoop (Token token, Token comma, Tokenizer stream, Node l) {
//    if (token == (stream.peekToken())) {
//      stream.readToken();
//      return (Node)(l.reverse());
//    } else if (comma == (stream.peekToken())) {
//      stream.readToken();
//      if (token == (stream.peekToken())) {
//	stream.readToken();	
//	return (Node)(l.reverse());
//      } else {
//	return prsmatchModifiedLoop(token, comma, stream,
//				    new Node(parse(10, stream), l));
//      }
//    } else {
//      throw new ParserException("#COMMA! " + stream.readToken());
//    }
//  }
//
  private Tree delimErr (Token token, Tokenizer stream) {
    System.err.println("delimErr");
    throw new ParserException("#SYNTAX!"  + token);
  }

  private Tree erbErr (Token token, Tree left, Tokenizer stream) {
    System.err.println("erbErr");
    throw new ParserException("#SYNTAX!"  + token);
  }

  private Tree premtermErr (Token token, Tokenizer stream) {
    System.err.println("premtermErr");
    throw new ParserException("#SYNTAX! " + token);
  }

  private Tree openParenNud (Token token, Tokenizer stream) {
    if ((stream.peekToken()) == Token.RPAR) {
      stream.readToken();
      return null;
    } else {
      return prsmatch(Token.RPAR, Token.COMMA, stream);
    }
  }

  private Tree openParenLed (Token token, Tree left, Tokenizer stream) {
    return new Node(((Token)left).getHeader(),
		    prsmatch(Token.RPAR, Token.COMMA, stream));
  }

  //  public static void main(String args[]) {
  //    String string = (args.length == 0) ? "1+2+3+A2+SQRT(4)" : args[0];
  //    System.err.println(new Pratt(null).toplevelParse(string));
  //  }

  final class Tokenizer implements Constants {
    Hashtable objects = new Hashtable();
    Hashtable strings = null;

    int pos;

    LinkedList l;

    final void setString(String string) {
      l = new LinkedList();
      pos = 0;

      // You might want to use StringTokenizer,
      // but it doesn't handle both ">=" and ">" well.
      int length = string.length();
      char prefix = 0;
      boolean inquote =  false;
      StringBuffer sb = new StringBuffer();
      for (int i = 0; i < length; i++) {
	char c = string.charAt(i);
	if (prefix != 0) {
	  switch(c) {
	  case '=':
	    // add last token
	    if (sb.length() != 0) {
	      l.add(intern(sb.toString()));
	      sb.setLength(0);
	    }
	    // add second token
	    {
	      sb.append(prefix);
	      sb.append(c);
	      prefix = 0;
	      l.add(intern(sb.toString()));
	      sb.setLength(0);
	    }
	    // Continue with next iteration of loop
	    continue;
	  default:
	    // otherwise add last token, prefix, and fall through for next
	    if (sb.length() != 0) {
	      l.add(intern(sb.toString()));
	      sb.setLength(0);
	    }
	    {
	      sb.append(prefix);
	      l.add(intern(sb.toString()));
	      sb.setLength(0);
	    }
	    prefix = 0;
	    // fall through to next switch
	  }
	}
	// fall through from above switch.
	if (inquote && c != '"') {
	  if (c == '\\' && i < length)
	    c = string.charAt(++i);
	  sb.append(c);
	} else {
	  switch(c) {
	  case ' ':
	  case '\t':
	    if (sb.length() != 0) {
	      l.add(intern(sb.toString()));
	      sb.setLength(0);
	    }
	    break;

	  case '"':
	    if (inquote) {
	      inquote = false;
	    } else {
	      inquote = true;
	    }
	    sb.append(c);
	    break;

	  case '=':
	    // Skip initial character "=" in formula.
	    if (i == 0) 
	      continue;
	    /** FALL THROUGH ***/
	  case '^':
	  case ',':
	  case '+':
	  case '-':
	  case '*':
	  case '/':
	  case '(':
	  case ')':
	  case ':':
	  case '%':
	    if (sb.length() != 0) {
	      l.add(intern(sb.toString()));
	      sb.setLength(0);
	    }
	    sb.append(c);
	    l.add(intern(sb.toString()));
	    sb.setLength(0);
	    break;

	  case '<':
	  case '>':
	  case '!':
	    prefix = c;
	    break;

	  default:
	    sb.append(c);
	  }
	}
      }
      if (prefix != 0)
	sb.append(prefix);
      if (sb.length() != 0) {
	l.add(intern(sb.toString()));
      }
      l.add(Token.EOD);
    }

    private void add(Token t) {
      objects.put(t.getSymbol(), t);
    }

    //  private Token def(String symbol, int lbp, int rbp, int nud, int led) {
    //    Token t = new Primitive(symbol, lbp, rbp, nud, led);
    //    add(t);
    //    return t;
    //  }

    Token intern(String string) {
      Token t;
      if (isQuoted(string)) {
	return internString(string);
      } else if (isNumber(string)) {
	return internNumber(string);
      } else if (isCellRef(string)) {
	return internCellRef(string);
      } else {
	return internSymbol(string);
      }
    }

    private boolean isQuoted(String string) {
      return (string.length() > 1 && string.charAt(0) == '"');
    }

    public boolean isNumber(String string) {
      int len = string.length();
      if (len < 1) return false;
      char c = string.charAt(0);
      if (Character.isDigit(c) || (c == '-') || (c == '.')) {
	try {
	  Token t = internNumber(string);
	  return true;
	} catch (Exception e) {
	  return false;
	}
      } else 
	return false;
    }

    private Token internString(String string) {
      // don't intern them with the other objects to avoid confusion
      Token result = null;
      if (strings == null)
	strings = new Hashtable();
      else {
	result = (Token)strings.get(string);
	if (result != null) 
	  return result;
      }
      result = new StringToken(string.substring(1, string.length()-1));
      strings.put(string, result);
      return result;
    }


    private Token internNumber(String string) {
      Token t = (Token)objects.get(string);
      if (t == null) {
	// int i = string.indexOf('.');
	//      if (i != -1 && i != string.length())
	//	t = new DoubleToken(string);
	//      else
	t = new BigDecimalToken(string);
	add(t);
      }
      return t;
    }

    /**
     * There are two main kinds of cell references,
     * each with some features.
     * <ul>
     * <li>Alphanumeric A1</li>
     * <li>Positional R[1]C[1]</li>
     * </ul>
     * 
     * To make matters more confusing, alphanumeric can be either relative
     * or absolute.  The normal-looking A1 style references are actually
     * relative to the current cell, and should be adjusted when copied
     * and pasted to another cell.  To have an sbsolute reference,
     * the alphanumeric style uses a $ prefix before the absolute
     * component (row or column).  So $A1 is column A absolute,
     * and A$1 is row 1 absolute, and $A$1 is both absolute.
     *
     * RC cell references are taken care of in the upper level CSV parser,
     * and converted to A1-style relative references before we see them here.
     *
     */
    private boolean isCellRef(String string) {
      int len = string.length();
      // Must be at least 2 characters.
      if (len < 2) 
	return false;
      int i = 0;
      {
	char c = string.charAt(i);
	if (c == '$') {
	  i++;
	  c = string.charAt(i);
	}
	if (! Character.isLetter(c)) return false;
	i++;
      }
      // If we fell off end, it's not a cell ref.
      if (i > len) return false;
      // Now validate col.
      boolean rowdollar = false;
      boolean rowseen = false;
      for (; i < len; i++) {
	char c = string.charAt(i);      
	if (c == '$') {
	  if (rowdollar) break;
	  rowdollar = true;
	  continue;
	}
	if (!Character.isDigit(c)) break;
	if (c != '0') rowseen = true;
      }
      if (i == len) return rowseen;
      else return false;
    }

    private Token internCellRef(String string) {
      Token t = (Token)objects.get(string);
      if (t == null) {
	if (string.indexOf('$') == -1)
	  t = new CellRef(string);
	else
	  t = new AbsoluteCellRef(string);	  
	add(t);
      }
      return t;
    }


    private Token internSymbol(String string) {
      Token t = (Token)objects.get(string);
      if (t == null) {
	t = new Symbol(string);
	add(t);
      }
      return t;
    }

    final Token readToken() {
      // not l.size now that it's a linkedlist
      if (pos == l.size()) return null;
      Token t = ((Token)l.get(pos++));
      return t;
    }

    final Token peekToken() {
      if (pos == l.size()) return null;
      Token t = ((Token)l.get(pos));
      return t;
    }

    Tokenizer() {
      if (Token.EOD == null) Token.EOD = new EOD();
      add(Token.EOD);
      add(new LPar());
      if (Token.RPAR == null) Token.RPAR = new RPar();
      add(Token.RPAR);
      add(new Colon());
      add(new Minus());
      add(new Plus());
      add(new Times());
      add(new Equals());
      add(new Exponent());
      add(new Divide());
      add(new Percent());
      add(new Less());
      add(new Greater());
      add(new GreaterEqual());
      add(new NotEqual());
      add(new LessEqual());
      add(new Not());
      add(new And());
      add(new Or());
      add(new If());
      add(new Sum());
      add(new Product());
      add(new Max());
      add(new Min());
      add(new Count());
      add(new Debug());
      add(new SetBackgroundColor());
      add(new SetForegroundColor());
      add(new LockRange());
      add(new RefError());
      if (Token.COMMA == null) Token.COMMA = new Comma();
      add(Token.COMMA);
      if (Token.ZERO == null)
	Token.ZERO = new BigDecimalToken(BigDecimal.ZERO);
      add(Token.ZERO);
      if (Token.ONE == null)
	Token.ONE = new BigDecimalToken(BigDecimal.ONE);
      add(Token.ONE);
      if (Token.TEN == null)
	Token.TEN = new BigDecimalToken(BigDecimal.TEN);
    };
  }
}
