package com.hipme.flipsheet;

import danger.app.Application;
import danger.util.LinkedList;
import danger.util.StringTokenizer;
import danger.util.StringUtils;

// Except for empty values, we always require quote (") quoted values, and use ("") for quoting a quote.
// We might move to dispensing with quotes, especially on numbers and formulas.
final class CSVParser /* implements Iterator */ {
  private static final boolean debugcsv = false;
  final String csv;
  final int length;
  String sheetName;
  LinkedList onLoad;
  int[] cellWidths;
  int cursorRow;
  int cursorCol;
  int currentRow = 1;
  int currentCol = 1;
  int precision = 9;
  int nRows;
  boolean isWheelAxisHorizontal = true;
  boolean newlineNext = false;

  int pos = 0;
  CSVParser (String csv) {
    this.csv = csv;
    this.length = csv.length();
    if (! parseComment()) {
      if (debugcsv) System.err.println("parsing initial comment failed; trying from pos=0");
      // if failed parsing comment try again as cell contents.
      pos = 0;
    }
  }

  // even if hasNext() is true, next might return "" or null anyway.
  public boolean hasNext() { 
    return pos < length || newlineNext;
  }


  public Object next() {
    if (newlineNext) {
      newlineNext = false;
      advanceRow();
      return null;
    } else {
      advanceCol();
    }
    
    boolean done = false;
    boolean inquote = false;
    boolean quotemode = false;
    StringBuffer sb = null;
    while (pos < length && ! done) {
      char c = csv.charAt(pos);
      switch(c) {
      case '"':
	pos++;
	// prolly in !quotemode we should have "" mean a literal " because that's what other seem to do.
	if (! inquote) {
	  inquote = true;
	  quotemode = true;
	} else {
	  inquote = false;
	  quotemode = false;
	  if (pos >= length) {
	    done = true;
	  }
	}
	break;

      case '\\':
	// \" is quoted quote.
	// this is broken when not in !quotemode
	pos++;
	if (inquote && pos<length) {
	  if (sb == null) sb = new StringBuffer();
	  sb.append(csv.charAt(pos));
	  pos++;
	} 
	break;

	// both newline terminates an entry outside strings and pushes a newline for next item.
      case '\n':
	if (inquote) {
	  if (sb == null) sb = new StringBuffer();
	  sb.append(c);
	} else {
	  done = true;
	  newlineNext = true;
	}
	quotemode = false;
	pos++;
	break;

	// comma terminates an entry outside strings
      case ',':
	if (inquote) {
	  if (sb == null) sb = new StringBuffer();
	  sb.append(c);
	} else {
	  done = true;
	  quotemode = false;
	}
	pos++;
	break;

	// return char we insert in quoted string or ignore completely
      case '\r':
	if (inquote) {
	  if (sb == null) sb = new StringBuffer();
	  sb.append(c);
	} 
	pos++;
	break;

      default:
	if (inquote || !quotemode) {
	  if (sb == null) sb = new StringBuffer();
	  sb.append(c);
	} else
	  System.err.println("flipsheet: Ignoring character " + c +" outside required quotes");
	pos++;
      }
    }
    if (sb == null) 
      return "";
    else {
      //
      String result = sb.toString();
      // String result2 = parseRCCellRef(result, currentRow, currentCol);
      //      if (result2 != null)
      //	return result2;
      //      else
	return result;
    }
  }

  private boolean parseComment() {
    // If the first line doesn't start with a '#' then just give up now.
    if (pos>=length || csv.charAt(pos) != '#') {
      if (debugcsv) System.err.println("flipsheet parseComment: 1; pos="+pos+" length="+length+" charAt(pos)="+csv.charAt(pos)+" csv="+csv);
      return false;
    }
    // The right thing to would be to store the metadata into a different DataStore
    // and use UIDs to refer to the other entries,
    // and so the amount of data updated when you just move the cursor
    // and don't change the sheet is lower.  But we just store
    // them as comments in the serialized CSV.
    while (pos < length) {
      // System.err.println("Looking at char " + csv.charAt(pos)+" at position " + pos + " length = " + length);
      // Skip CR or LF (sometimes we get both)
      {
	char c = csv.charAt(pos);
	if ((c == '\r') || (c == '\n')) {
	  pos++;
	  continue;
	}
      }
      if (csv.charAt(pos) != '#') {
	if (debugcsv) System.err.println("flipsheet parseComment: 2 " +((int)csv.charAt(pos)));
	return true;
      }
      if (++pos >= length) {
	if (debugcsv) System.err.println("flipsheet parseComment: 3");
	return true;
      }
      switch(csv.charAt(pos)) {
	// #<space>[|-]A42<space>My Spreadsheet Title<newline>
      case ' ': {
	if (++pos >= length) {
	  if (debugcsv) System.err.println("flipsheet parseComment: 4");
	  return false;
	}
	// #<space>A42<space>My Spreadsheet Title<newline>
	//         P
	// #<space>[-|]A42<space>My Spreadsheet Title<newline>
	//         P
	char in = csv.charAt(pos);
	if (in == '-') {
	  isWheelAxisHorizontal = true;
	  pos++;
	} else if (in == '|') {
	  isWheelAxisHorizontal = false;
	  pos++;
	} 

	if (pos >= length) {
	  if (debugcsv) System.err.println("flipsheet parseComment: 5");
	  return false;
	}
	in = Character.toUpperCase(csv.charAt(pos++));
	if ('A' <= in && in <= (((char)SheetTable.MAX_COLS)+'@')) {
	  cursorCol = (in - 'A') + 1;
	} else {
	  if (debugcsv) System.err.println("flipsheet parseComment: 6: "+ in);
	  return false;
	}

	// #<space>A42<space>My Spreadsheet Title<newline>
	//          P
	if (pos >= length) {
	  if (debugcsv) System.err.println("flipsheet parseComment: 7");;
	  return false;
	}
	int i = csv.indexOf(' ', pos);	  // skip over the space.
	if (i == -1) {
	  if (debugcsv) System.err.println("flipsheet parseComment: 8");
	  return false;
	}
	// #<space>A42<space>My Spreadsheet Title<newline>
	//          P I
	try {
	  if (pos >= length) {
	    if (debugcsv) System.err.println("flipsheet parseComment: 9");
	    return false;
	  }
	  cursorRow = StringUtils.toInt(csv.substring(pos, i));
	  pos = i+1;
	} catch (NumberFormatException e) {
	  if (debugcsv) System.err.println("flipsheet parseComment: 9.5");
	  if (debugcsv) e.printStackTrace();
	  return false;
	}
	// #<space>A42<space>My Spreadsheet Title<newline>
	//            I      P
	StringBuffer sb = new StringBuffer();
	while(pos < length && (csv.charAt(pos) != '\r' && csv.charAt(pos) != '\n')) {
	  sb.append(csv.charAt(pos));
	  pos++;
	  if (pos >= length) {
	    if (debugcsv) System.err.println("flipsheet parseComment: 10");
	    return false;
	  }
	}
	skipEOL();
	sheetName = sb.toString();
	break;
      }

	// Set number of rows and columns.
	// First number is number of rows.
	// Remaining numbers are pixel widths of columns.  First column is heading.
	// Ignore and use defaults if there's a problem.
      case '#': {
	pos++;
	int eol = findEOL();
	if (eol == -1) eol = csv.length();
	String data = csv.substring(pos, eol);
	if (debugcsv) System.err.println("##" + data);
	int[] cw = null;
	int nr;
	{
	  int i = 0;
	  StringTokenizer st = new StringTokenizer(data, " \r\n");
	  while (st.hasMoreTokens()) {
	    st.nextToken();
	    i++;
	  }
	  // Must have at least 1 row number 2 columns
	  int nc = i - 1;
	  if (nc >= 2)
	    cw = new int[nc];
	}
	{
	  int i = 0;
	  int w = 0;

	  StringTokenizer st = new StringTokenizer(data, " \r\n");
	  nr = StringUtils.toInt(st.nextToken().trim());
	  try {
	    while (st.hasMoreTokens()) {
	      int val = StringUtils.toInt(st.nextToken().trim());
	      w += val;
	      cw[i++] = val;
	    }
	  } catch (NumberFormatException fe) {
	    if (debugcsv) System.err.println("flipsheet: ## " + fe);
	    cw = null;
	  }
	  if ((w != SheetTable.TABLE_WIDTH) && ((flipsheet)Application.getCurrentApp()).exactTableWidth) {
	    SheetWindow.errorprint("table width " + w + "!="+SheetTable.TABLE_WIDTH+"; ignoring");
	    cw = null;
	  }
	  if (nr > SheetTable.MAX_ROWS) {
	    SheetWindow.errorprint("nRows="+nr+">"+SheetTable.MAX_ROWS+"; ignoring");
	    cw = null;
	  }
	}

	// If all went OK, set the width and rows.
	if (cw != null) {
	  cellWidths = cw;
	  nRows = nr;
	}
	skipEOL();
	break;
      }

	// Look for optional lines starting with '#=', which are expressions to execute after load.
	// (Though they really ought to have dependencies).
      case '=': {
	pos++;
	int eol = findEOL();
	if (onLoad == null) onLoad = new LinkedList();
	String expr = csv.substring(pos, eol).trim();
	if (expr.length() > 0) {
	  onLoad.add(expr);
	}
	skipEOL();
	break;
      }

	// #.precision
      case '.': {
	pos += 1;
	int eol = findEOL();
	if (onLoad == null) onLoad = new LinkedList();
	String p = csv.substring(pos, eol).trim();
	if (p.length() > 0) {
	  // System.err.println("precision = '"+p+"' pos="+pos+" eol="+eol);
	  precision = StringUtils.toInt(p);
	}
	skipEOL();
	// System.err.println("precision set to " + precision);
	break;
      }

      default: {
	System.err.println("did not understand char " + csv.charAt(pos) + " at column " + pos);
	skipEOL();
	break;
      }
      }
    }
    return true;
  }

  // eat through \r or \n left over
  private void skipEOL() {
    pos = findEOL()+1;
    if (pos > length) pos = length;
  }

  private int findEOL() {
    int i = pos;
    char c;
    while((i < length) && 
	  ((c = csv.charAt(i)) != '\r') && 
	  (c != '\n')) {
      i++;
    }
    return i;
  }


  private void advanceRow() {
    currentCol = 1;
    currentRow++;
  }
  private void advanceCol() {
    currentCol++;
  }


//  /**
//   * For R[1]C, R[1]C[2], RC[2], R[-1]C, R[-1]C[-2], RC[-2], etc.
//   * If ref is not of that form, return null.
//   * Else return String of an A1-style relative ref relative to currentRow and currentColumn.
//   * Omitted brackets are equivalent to 0 offset.
//   * So RC[1] is one column over to the right from the currentCol
//   * in current row, and R[2]C[-1] is two rows down
//   * and one column to the left of currentRow.  I suppose
//   * RC would be the current cell, but that's not too useful anyway,
//   * so we can ignore it.  So the minimum length is "RC[1]" for 5 chars,
//   * and it always starts with an R (or r?)
//   * so we can use that as a cheap test.
//   */
//  private String parseRCCellRef(String string) {
//    boolean ok = false;
//    int row = 0;
//    int col = 0;
//    // Must be at least 5 characters
//    if (len < 5 || string.charAt(0) != 'R') 
//      return null;
//    int i = 1;
//    while (!ok) {
//      // R*
//      char c = string.charAt(i);
//      // must be either [ or C.	
//      if (c == '[') {
//	// R[
//	// Expect a positive or negative rowDelta followed by ']'
//	if (++i >= len) return null;
//	{
//	  int[] r = StringUtils.toIntNextIndex(string, i);
//	  deltaRow = r[0];
//	  i = r[1];			// end of number
//	}
//	if (i >= len) return null;
//	c = string.charAt(i);	// expecting ']';
//	if (c != ']') 
//	  return null;
//	if (++i >= len) return null;
//	c = string.charAt(i);
//	if (c != 'C') 
//	  return null;
//	// R[#]C
//	// followed by EOD or '[' -? # ']'
//	if (++i >= len) {
//	  ok = true;
//	  break;
//	}
//	c = string.charAt(i);
//	if (c == '[') {
//	  // R[#]C[
//	  if (++i >= len)
//	    return null;
//	}
//	// R[#]C[.
//	c = string.charAt(i);
//	if (c == '-' || Character.isDigit(c)) {
//	  // R[#]C[#
//	  int[] r = StringUtils.toIntNextIndex(string, i);
//	  deltaCol = r[0];
//	  i = r[1];			// end of number
//	  if (i >= len) 
//	    return null;
//	  c = string.charAt(i);
//	  if (c != ']')
//	    return null;
//	  ok = true;
//	  break;
//	} else {
//	  return null;
//	}
//      } else if (c == 'C') {
//	// RC
//	i++;
//	// Expect for '[' 
//	if (string.charAt(i) != '[') 
//	  return null;
//	// RC[
//	// followed by -? # ']'
//	if (c == '-' || Character.isDigit(c)) {
//	  deltaCol = StringUtils.toInt(string, i);
//	} else {
//	  return null;
//	}
//	ok = true;
//	break;
//      } else {
//	return null;
//      }
//    }
//    return (((char)((currentCol+deltaCol+'@'))) + (currentRow+deltaRow));
//  }
}
