package com.hipme.flipsheet;

import danger.app.*;
import danger.internal.Date;
import danger.ui.*;
import danger.util.LinkedList;
import danger.util.ListIterator;
import danger.util.StringTokenizer;
import danger.system.*;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;
import danger.util.DEBUG;
import danger.util.format.StringFormat;

import com.hipme.flipsheet.parser.Pratt;
import com.hipme.flipsheet.parser.ParserException;

public final class SheetTable extends Table implements Resources, Events {
  // 128 rows header + A1:A127
  static final int MAX_ROWS = 128; 
 // 26 Columns A:Z
  static final int MAX_COLS = 26;
  static final int DEFAULT_ROWS = MAX_ROWS / 2;
  static final int SCREEN_WIDTH = 240; // EVIL
  static final int SCROLL_BAR_WIDTH = 3; // EVIL
  static final int TABLE_WIDTH = SCREEN_WIDTH-SCROLL_BAR_WIDTH;
  static final int[] DEFAULT_CELL_WIDTHS = { 12, 55, 55, 55, 60};
  private static int HEADER_BACKGROUND_COLOR = Style.getSystemStyle().getColor(Style.TITLE_BAR_PILL_COLOR);
  private static int HEADER_FOREGROUND_COLOR = Style.getSystemStyle().getColor(Style.TITLE_BAR_TEXT_COLOR);

  static int defaultBackgroundColor = Color.WHITE;
  static int highlightBackgroundColor = Color.YELLOW;

  // nRows is 0-based so the header row counts as 1!
  static int nRows = DEFAULT_ROWS;
  int[] cellWidths = DEFAULT_CELL_WIDTHS;
  int nCols = cellWidths.length;
  int cellHeight;
  Font valueFont;
  Font labelFont;
  String sheetName;
  int rowCursor = 1;
  int colCursor = 1;
  boolean wheelAxisHorizontal = true;
  Pratt parser;
  LinkedList masterDependencies;
  LinkedList onLoad=null;
  Rect tmpRect = new Rect();	// good idea or not?
  /**
   * If true, take descenders into account to "g" and "p" look nice,
   * lose a couple of lines of the screen.
   */
  boolean showDescenders = false;
  
  /**
   * If true, outline rather than use background.
   */
  boolean useOutline = false;

  public SheetTable(String csv) {
    this(csv, getDefaultSheetName(getFlipsheet().getSheetNumber()));
  }

  public SheetTable(String csv, String sheetName) {
    super();
    // DEBUG.p("initializing " + sheetName + " from " + csv);
    initialize(csv, sheetName);
  }

  int getRowCursor() {return rowCursor;}
  int getColCursor() {return colCursor;}

  static String getDefaultSheetName(int i) {
    return StringFormat.fromRsrc(kStrSheetFormat, String.valueOf(i));
  }

  String getSheetName() {
    return sheetName;
  }

  void setSheetName(String sheetName) {
    this.sheetName=sheetName;
  }

  String getCellName() {
    // cache this?
    return ((char)('A' + (colCursor - 1))) + "" + (rowCursor);
  }

  String getOnLoadText() {
    if (onLoad == null) return "";
    StringBuffer sb = new StringBuffer();
    for (ListIterator it = onLoad.listIterator(); it.hasNext(); ) {
      sb.append(it.next());
      sb.append("\n");
    }
    return sb.toString();
  }

  void setOnLoadText(String text) {
    onLoad = null;
    addOnLoadText(text);
  }

  /**
   * For now, all this does is append a series of expressions.
   * Some day it should merge ranges to eliminate duplicates
   * and flip/flops.
   */
  void addOnLoadText(String text) {
    StringTokenizer st = new StringTokenizer(text, "\r\n");
    while (st.hasMoreTokens()) {
      addOnLoadExpression(st.nextToken());
    }
  }

  /**
   * For now, all this does is append one expression
   * Some day it should merge ranges to eliminate duplicates
   * and flip/flops.
   */
  void addOnLoadExpression(String expr) {
    try {
      String value = evaluateToObject(expr).toString();
      if (onLoad == null) 
	onLoad = new LinkedList();
      onLoad.addLast(expr);
    } catch (Exception e) {
      SheetWindow.errorprint(e, "onLoad expr " + expr + " got error ");
    }
  }

  void nextRow() { 
    setRowCursor(rowCursor+1); 
  }

  void nextCol() { 
    setColCursor(colCursor+1); 
  }

  void previousRow() { 
    setRowCursor(rowCursor-1); 
  }

  void previousCol() { 
    setColCursor(colCursor-1); 
  }

  /**
   * Moves forward one cell.  Wraps to previous or next row
   * at ends of line.  Returns true if wrapped, false if not.
   */
  boolean forwardCell() {
    if (colCursor == nCols-1) {
      setCursor(rowCursor+1, 1);
      return true;
    } else {
      nextCol();
      return false;
    }
  }

  /**
   * Moves back one cell.  Wraps to previous or next row
   * at ends of line.  Returns true if wrapped, false if not.
   */
  boolean backCell() {
    if (colCursor == 1) {
      setCursor(rowCursor-1, nCols-1);
      return true;
    } else {
      previousCol();
      return false;
    }
  }

  void setCursor(int row, int col) {
    if (row < 1 || col < 1) return;
    if (row >= getRows() || col >= getColumns()) return;
    clearHighlighted(rowCursor, colCursor);
    rowCursor = row;
    colCursor = col;
    setHighlighted(rowCursor, colCursor);
  }

  void setRowCursor(int i) {
    if (i < 1) return;
    if (i >= getRows()) return;
    clearHighlighted(rowCursor, colCursor);
    rowCursor = i;
    setHighlighted(rowCursor, colCursor);
  }

  void setColCursor(int i) {
    if (i < 1) return;
    if (i >= getColumns()) return;
    clearHighlighted(rowCursor, colCursor);
    colCursor = i;
    setHighlighted(rowCursor, colCursor);
  }

  void clearHighlighted(int row, int col) {
    SheetTableCell td = ((SheetTableCell)getTableCell(row, col));
    if (td != null) {
      td.setHighlighted(false);
      getCellBounds(row, col, tmpRect);
      if (useOutline) includeCellBorder(tmpRect);
      invalidate(tmpRect);
    } else {
      //SheetWindow.errorprint("setCellBackgroundColor " + row +","+col+ " was null");
    }
  }

  void setHighlighted(int row, int col) {
    SheetTableCell td = ((SheetTableCell)getTableCell(row, col));
    if (td != null) {
      td.setHighlighted(true);
      getCellBounds(row, col, tmpRect);
      if (useOutline) includeCellBorder(tmpRect);
      invalidate(tmpRect);
    } else {
      //SheetWindow.errorprint("setCellBackgroundColor " + row +","+col+ " was null");
    }
  }

  public void setCellBackgroundColor(int row, int col, int color) {
    SheetTableCell td = (SheetTableCell)getTableCell(row, col);
    if (td != null) {
      td.backgroundColor=color;
      td.setBackgroundColor(color);
      getCellBounds(row, col, tmpRect);
      invalidate(tmpRect);
    } else {
      //SheetWindow.errorprint("setCellBackgroundColor " + row +","+col+ " was null");
    }
  }

  public void setCellForegroundColor(int row, int col, int color) {
    SheetTableCell td = (SheetTableCell)getTableCell(row, col);
    if (td != null) {
      td.foregroundColor=color;
      td.setForegroundColor(color);
      getCellBounds(row, col, tmpRect);
      invalidate(tmpRect);
    } else {
      // SheetWindow.errorprint("setCellForegroundColor " + row +","+col+ " was null");
    }
  }

  Rect getFastCellBounds(Rect rect) {
    return getFastCellBounds(rowCursor, colCursor, rect);
  }

  /**
   * Get only the Y portion of the bounds and leave the X as table bounds,
   * for use with ScrollView, which only cares about the Y.
   */
  Rect getFastCellBounds(int row, int col, Rect rect) {
    // fill in to the size of the table to get the X.
    getBounds(rect);
    int borderWidth = getBorderWidth();
    int prevBot = borderWidth;
    for (int j = 0; j <= row; j++) {
      int height = getTableCell(j, 0).getHeight();
      rect.top = prevBot;
      rect.bottom = rect.top + height;
      prevBot = prevBot + height + borderWidth;
    }
    return rect;
  }

  /**
   * This method returns the cell bounds accurate
   * in both X and Y directions, but is more expensive
   * to compute.
   */
  Rect getCellBounds(int row, int col, Rect rect) {
    getBounds(rect);
    int borderWidth = getBorderWidth();
//
// int prevBot = borderWidth;
// for (int j = 0; j <= row; j++) {
//	int height = getTableCell(j, 0).getHeight();
//	rect.top = prevBot;
//	rect.bottom = rect.top + height;
//	prevBot = prevBot + height + borderWidth;
// }
//
    {
      int prevBot = borderWidth;
      int height = getTableCell(0, 0).getHeight();
      for (int j = 0; j <= row; j++) {
	rect.top = prevBot;
	rect.bottom = rect.top + height;
	prevBot = prevBot + height + borderWidth;
      }
    }
    {
      int prevLeft = borderWidth;
      for (int i = 0; i <= col; i++) {
	int width = getTableCell(row, i).getWidth();
	rect.left = prevLeft;
	rect.right = rect.left + width;
	prevLeft = prevLeft + width + borderWidth;
      }
    }
    return rect;
  }

  Rect getCellBounds(Rect rect) {
    return getCellBounds(rowCursor, colCursor, rect);
  }

  Rect includeCellBorder(Rect rect) {
    rect.top -= 1;
    rect.left -= 1;
    rect.right += 2;
    rect.bottom += 2;
    return rect;
  }


  /**
   * ScrollView will sometimes scroll to make sure this rect is in view,
   * so we might as well use it ourselves.
   */
  public Rect getInterestingRect(Rect rect) {
    // If we're in Column 1, then be sure to show column 0 too!
    if (colCursor == 1) {
      // actually we should merge them, but if header and Col 1 won't fit
      // on the screen, we've got bigger problems.
      return getCellBounds(rowCursor, 0, rect);
    } 
    else {
      return /*getFastCellBounds*/getCellBounds(rect);
    }
  }

  void setHighlightBackgroundColor(int color) {
    highlightBackgroundColor = color;
  }

  boolean isWheelAxisHorizontal() {
    return wheelAxisHorizontal;
  }

  void setWheelAxisHorizontal(boolean w) {
    wheelAxisHorizontal = w;
    setCellContents(0, 0, wheelAxisHorizontal ? "-" : "|");
    getCellBounds(0, 0, tmpRect);
    invalidate(tmpRect);
  }

  public boolean addRow(Vector vector) {
    boolean value = super.addRow(vector);
    if (getRows() == rowCursor - 1)
      setCellBackgroundColor(getRowCursor(), getColCursor(), highlightBackgroundColor);
    return value;
  }

  TableCell getCell() {
    return getTableCell(getRowCursor(), getColCursor());
  }

  String getCellContents() {
    return getCellContents(getRowCursor(), getColCursor());
  }

  String getCellAsString() {
    return getCell().getElementAsString();
  }

  /**
   * Returns width in pixels -- assumes value cell not header
   * as the header font is different.
   */
  int getCellStringWidth() {
    return valueFont.getWidth(getCell().getElementAsString());
  }


  String getCellContents(int row, int col) {
    TableCell td = getTableCell(row, col);
    if (td == null) return null;
    if (td instanceof SheetTableCell) {
      return ((SheetTableCell)td).getContents();
    } else {
      return td.getElementAsString();
    }
  }

  void setCellContents(String value) {
    setCellContents(getRowCursor(), getColCursor(), value);
  }

  void setCellContents(int row, int col, String newValue) {
    TableCell td = getTableCell(row, col);
    if (td instanceof SheetTableCell) {
      ((SheetTableCell)td).setCellContents(newValue);
    } else {
      td.setElementToString(newValue);
    }
  }

  boolean isCellLocked() {
    return ((SheetTableCell)getCell()).isLocked();
  }

  boolean isCellLocked(int row, int col) {
    TableCell td = getTableCell(row, col);
    if (td instanceof SheetTableCell) {
      return ((SheetTableCell)td).isLocked();
    } else {
      return true;
    }
  }

  public void setLocked(boolean locked) {
    ((SheetTableCell)getCell()).setLocked(locked);
  }

  public void setLocked(int row, int col, boolean locked) {
    SheetTableCell td = (SheetTableCell)getTableCell(row, col);
    // Will be null of lock is called on a range that is larger than the table.
    if (td != null) 
      td.setLocked(locked);
  }

  /**
   * Normally we show descenders like "y" but to get 
   * several more rows on the screen you can turn that off
   */
  void setShowDescenders(boolean showDescenders) {
    if (this.showDescenders == showDescenders) return;
    this.showDescenders = showDescenders;

    calculateCellHeight();
    for (int row = 0; row < nRows; row++) {
      for (int col = 0; col < nCols; col++) {
	getTableCell(row, col).setSize(cellWidths[col], cellHeight);
      }
    }
    invalidate();
  }

  boolean getShowDescenders() {
    return showDescenders;
  }

  /**
   * Use outline or highlight background color fur current cell.
   */
  void setUseOutline(boolean useOutline) {
    if (this.useOutline == useOutline) return;
    clearHighlighted(rowCursor, colCursor);
    this.useOutline = useOutline;
    setHighlighted(rowCursor, colCursor);
  }

  boolean getUseOutline() {
    return useOutline;
  }

  void calculate() {
    if (masterDependencies == null) {
      long t = System.currentTimeMillis();
      masterDependencies = calculateMasterDependencies();
      DEBUG.p("Calculating masterDependencies took " + (System.currentTimeMillis()-t) + " ms");
    }
    if (masterDependencies == null) {
      DEBUG.p("Skipping calculations because of circular dependencies");
      return;
    }
    {
      long t = System.currentTimeMillis();
      for (ListIterator it = masterDependencies.listIterator(); it.hasNext(); ) {
	SheetTableCell td = (SheetTableCell)it.next();
	td.calculate();
      }
      DEBUG.p("Calculating only required values took " + (System.currentTimeMillis()-t) + " ms");
    }
  }

  void invalidateMasterDependencies() {
    masterDependencies = null;
  }

  LinkedList calculateMasterDependencies() {
    LinkedList q = new LinkedList();
    LinkedList masterDependencies = new LinkedList();
    Hashtable degrees = new Hashtable();
    int num_vertex = 0;
    // Add all cells with inDegree=0 to the queue.
    for (int j = 1; j < nRows; j++) {
      for (int i = 1; i < nCols; i++) {
	SheetTableCell v = (SheetTableCell)getTableCell(j, i);
	if (v.getFormula() != null || v.getDependents() != null) {
	  num_vertex++;
	  LinkedList dependencies = v.getDependencies();
	  if (dependencies == null || dependencies.size() == 0) {
	    q.addLast(v);
	  } else {
	    degrees.put(v, new Integer(dependencies.size()));
	  }
	}
      }
    }
    while (q.size() > 0) {
      SheetTableCell v = (SheetTableCell)(q.removeFirst());
      masterDependencies.addLast(v);
      LinkedList dependents = v.getDependents();
      if (dependents != null) {
	ListIterator it = dependents.listIterator();

	while (it.hasNext()) {
	  SheetTableCell w = ((SheetTableCell)it.next());
	  Integer d1 = (Integer)degrees.get(w);
	  if (d1 != null) {
	    int deg = d1.intValue()-1;
	    if (deg == 0) {
	      degrees.remove(w);	// double check
	      q.addLast(w);
	    } else {
	      degrees.put(w, new Integer(deg));
	    }
	  } else {
	    DEBUG.p("For cell " + v + " the dependency " + w + " was not in the table");
	  }
	}
      }
    }
    if (masterDependencies.size() > num_vertex) {
      DEBUG.p("circular dependency: masterDependencies.size = " + masterDependencies.size() +
	      " num_vertex = " + num_vertex);
    }
    if (true) {
      if (degrees.size() != 0) {
	java.util.Enumeration e = degrees.keys();
	StringBuffer sb = new StringBuffer("Circular references: "); // I18N
	while (e.hasMoreElements()) {
	  SheetTableCell k = (SheetTableCell)(e.nextElement());
	  sb.append(k.toString());
	  sb.append(" ");
	  // not quite right because current cell bg cleared before cursor moves on.
	  k.setBackgroundColor(Color.RED);
	  invalidate(getCellBounds(k.row, k.col, tmpRect));
	}
	getFlipsheet().showCircularDependency(sb.toString());
	masterDependencies = null;
      }
      if (q.size() != 0)
	throw new RuntimeException("BUG: q.size != 0: " + degrees.size());
    }
    return masterDependencies;
  }

  static String plist(LinkedList list) {
    if (list == null) return "";
    StringBuffer sb = new StringBuffer();
    for (ListIterator it = list.listIterator(); it.hasNext(); ) {
      sb.append(it.next());
      sb.append(" ");
    }
    return (sb.toString());
  }

  void clear() {
    for (int col = 1; col < nCols; col++) {
      for (int row = 1; row < nRows; row++) {
	SheetTableCell td = (SheetTableCell)getTableCell(row, col);
	td.setCellContents(null);
	td.setForegroundColor(Color.BLACK);
	td.setBackgroundColor(Color.WHITE);
      }
    }
    setWheelAxisHorizontal(true);
    setCursor(1, 1);
    onLoad=null;
    invalidate();		// necessary?
  }

  void initialize(String csv, String sheetName) {
    Vector rowVec;
    final int borderWidth = 1;
    final int cellSpacing = 1;

    this.sheetName = sheetName;

    useOutline = getFlipsheet().useOutline;
    // Font.findFont("Fixed6x10");
    valueFont = Font.findSystemFont();
    labelFont = Font.findBoldSystemFont();
    calculateCellHeight();
    setPosition(0, 0);
    setLeft(0);
    setBorderWidth(borderWidth);
    setCellSpacing(cellSpacing);
    setBackgroundColor(defaultBackgroundColor);
    setBorderColor(Color.GRAY11);
    setSpacingColor(Color.GRAY11);

    // Parse the CSV comment, if any, to get properties
    // that must be set before table is created.
    CSVParser parser = null;
    // DEBUG.p("csv="+csv);
    if (csv != null && csv.length() > 0) {
      parser = new CSVParser(csv);
      setPrecision(parser.precision);
      if (parser.cellWidths != null) {
	nRows = parser.nRows;
	cellWidths = parser.cellWidths;
	nCols = cellWidths.length;
	DEBUG.p("setting sheet to " + nRows + "x" + nCols);
      } else {
	nRows = DEFAULT_ROWS;
	cellWidths = DEFAULT_CELL_WIDTHS;
      }
      nCols = cellWidths.length;
    }
    // Make column headings in row 0
    {
      rowVec = new Vector();
      for (int col = 0; col < nCols; col++) {
	TableCell td = new TableCell(this);
	td.setSize(cellWidths[col], cellHeight);
	td.setForegroundColor(HEADER_FOREGROUND_COLOR);
	td.setBackgroundColor(HEADER_BACKGROUND_COLOR);
	if (col == 0) {
	  td.setElementToString("");
	} else {
	  td.setElementToString("" + ((char)('A' + col - 1)));
	}
	td.setFont(labelFont);
	td.setAlignment(StaticText.HALIGN_CENTER);
	rowVec.addElement(td);
      }
      addRow(rowVec);
    }

    // Make row headings and fill table 
    for (int row = 1; row < nRows; row++) {
      rowVec = new Vector();
      for (int col = 0; col < nCols; col++) {
	if (col == 0) {
	  TableCell td = new TableCell(this);
	  td.setSize(cellWidths[col], cellHeight);
	  td.setForegroundColor(HEADER_FOREGROUND_COLOR);
	  td.setBackgroundColor(HEADER_BACKGROUND_COLOR);
	  td.setElementToInt(row);
	  td.setFont(labelFont);
	  td.setAlignment(StaticText.HALIGN_RIGHT);
	  rowVec.addElement(td);
	} else {
	  // Really should make these sparse instead.
	  SheetTableCell td = new SheetTableCell(this, row, col);
	  td.setSize(cellWidths[col], cellHeight);
	  td.setForegroundColor(Color.BLACK);
	  // MyTableCell defaults to System Font to save space.
	  // since TableCell provides no way to get the font.
	  // So this gets ignored for now and is always findSystemFont.
	  // td.setFont(valueFont);
	  rowVec.addElement(td);
	}
      }
      addRow(rowVec);
    }

    // Autosize widget
    autoSize();

    // If CSV is available, fill table now that
    // all the cells exist.
    // DEBUG.p("csv="+csv);
    if (parser != null) {
      long t = System.currentTimeMillis();
      boolean done = false;
      for (int row = 1; row < nRows && !done; row++) {
	for (int col = 1; ; col++) {
	  if (!parser.hasNext()) {
	    done = true;
	    break;
	  }
	  String v = (String)parser.next();
	  if (v == null) {
	    // null means early end of row (no more commas)
	    break;
	  }
	  SheetTableCell td = (SheetTableCell)getTableCell(row, col);
	  // DEBUG.p("Setting row="+row+" col="+col+" to " + v);
	  td.setCellContents(v);
	}
      }
      DEBUG.p("Setting values from CSV took " + (System.currentTimeMillis() - t) + "ms");
      // Check for onLoad
      onLoad=parser.onLoad;
      if (onLoad != null) {
	for (ListIterator it = onLoad.listIterator(); it.hasNext(); ) {
	  String expr = (String)it.next();
	  try {
	    DEBUG.p(expr + "->" + evaluateToObject(expr).toString());
	  } catch (Exception e) {
	    DEBUG.p("onload expr = " + expr + " got error " + e);
	  }
	}
      } 
    }

    // Restore or initialize table parameters
    if (parser == null) {
      setWheelAxisHorizontal(true);
      setCursor(1, 1);
    } else {
      setWheelAxisHorizontal(parser.isWheelAxisHorizontal);
      if (parser.currentRow != 0)
	setCursor(parser.currentRow, parser.currentCol);
      else
	setCursor(1, 1);
      if (parser.sheetName != null)
	this.sheetName = parser.sheetName;
      calculate();
    }
  }

  void calculateCellHeight() {
    showDescenders = getFlipsheet().showDescenders;
    cellHeight = valueFont.getAscent();
    if (showDescenders)
      cellHeight += valueFont.getDescent();
  }


  String toCSV() {
    return toCSV(0, 0, false, false);
  }

  /**
   * Returns empty string if empty.
   * @param rowDelta -- if positive, a number of blank rows to add after current row.
   * if negative, a number of rows to discard at current row, includign current row.
   */
  String toCSV(int rowDelta, int colDelta, boolean insertOnLeft, boolean makeTableWider) {
    StringBuffer sb = new StringBuffer();
    int lastCol = nCols;
    int lastRow = 0;
    int insertAtCol = insertOnLeft ? colCursor-1 : colCursor;
    // calculate lastRow.
    // BUG: This calculation fails to take int account the colDelta
    // and rowDelta so if the row or column that contributes
    // to lastRow is discarded, there will be too many rows/cols.
    boolean cellSeen = false;
    {
      for (int row = nRows-1; !cellSeen && (row > 0); row--) {
	for (int col = nCols-1; col > 0; col--) {
	  SheetTableCell td = (SheetTableCell)getTableCell(row, col);
	  String val = ((td == null) ? null : td.getContents());
	  if (val != null && val.length() > 0) {
	    cellSeen = true;
	    // don't change lastCol because we depend on having 
	    // a constant number of columns.
	    lastRow = row+1;
	    break;
	  } else {
	    // skip empty
	  }
	}
      }
    }
    
    // # !A1 Title
    outputTitle(sb);
    // ## 64 12 55 55 55 60
    outputCellWidths(sb, rowDelta, colDelta, insertOnLeft, makeTableWider);
    // #.9
    outputPrecision(sb);
    // #=expr1    
    outputOnLoad(sb, rowDelta, colDelta, insertOnLeft);

    // if there's no cells and no onLoad and we're not modifying the table,
    // consider it empty and stop here.
    if (!cellSeen && onLoad == null && rowDelta == 0 && colDelta == 0) 
      return "";

    int lastCol1 = lastCol-1; 
    // What follows are the normal, discard/insert row, and discard/insert column serialization cases.
    if (rowDelta == 0 && colDelta == 0) {
      // normal
      for (int i = 1; i < lastRow; i++) {
	for (int j = 1; j < lastCol; j++) {
	  SheetTableCell td = (SheetTableCell)getTableCell(i, j);
	  String contents = td.getContents();
	  if (contents == null || contents.length() == 0) {
	    sb.append((j != lastCol1) ? (",") : ("\n"));
	  } else {
	    sb.append("\"");
	    outputContents(sb, contents);
	    sb.append((j != lastCol1) ? ("\",") : ("\"\n"));
	  }  
	}
      }
    } else if (rowDelta < 0) {
      DEBUG.p("discard rows " + rowDelta);
      // discard rows
      for (int i = 1; i < lastRow; i++) {
	if (i == rowCursor) {
	  // We're deleting this -rowDelta rows, including this row.
	  int d = -rowDelta;
	  i += d - 1;
	  continue;
	}
	for (int j = 1; j < lastCol; j++) {
	  SheetTableCell td = (SheetTableCell)getTableCell(i, j);
	  String contents = td.getContents();
	  if (contents == null || contents.length() == 0) {
	    sb.append((j != lastCol1) ? (",") : ("\n"));
	  } else {
	    sb.append("\"");
	    String formula = td.getFormula();
	    if (formula == null) { 
	      outputContents(sb, contents);
	    } else {
	      outputContents(sb, reposition(formula, rowDelta, rowCursor, colDelta, colCursor));
	    }
	    sb.append((j != lastCol1) ? ("\",") : ("\"\n"));
	  }  
	}
      }
    } else if (rowDelta > 0) {
      DEBUG.p("insert rows " + rowDelta);
      // insert rows
      for (int i = 1; i < lastRow; i++) {
	if (i == rowCursor) {
	  for (int j = 1; j < lastCol; j++) {
	    sb.append((j != lastCol1) ? (",") : ("\n"));
	  }	  
	}
	for (int j = 1; j < lastCol; j++) {
	  SheetTableCell td = (SheetTableCell)getTableCell(i, j);
	  String contents = td.getContents();
	  if (contents == null || contents.length() == 0) {
	    sb.append((j != lastCol1) ? (",") : ("\n"));
	  } else {
	    sb.append("\"");
	    String formula = td.getFormula();
	    if (formula == null) { 
	      outputContents(sb, contents);
	    } else {
	      outputContents(sb, reposition(formula, rowDelta, rowCursor, colDelta, colCursor));
	    }
	    sb.append((j != lastCol1) ? ("\",") : ("\"\n"));
	  }
	}
      }
    } else if (colDelta < 0) {
      DEBUG.p("discard columns " + colDelta);
      // discard columns
      for (int i = 1; i < lastRow; i++) {
	for (int j = 1; j < colCursor; j++) {
	  SheetTableCell td = (SheetTableCell)getTableCell(i, j);
	  String contents = td.getContents();
	  if (contents == null || contents.length() == 0) {
	    sb.append((j != lastCol1) ? (",") : ("\n"));
	  } else {
	    sb.append("\"");
	    String formula = td.getFormula();
	    if (formula == null) { 
	      outputContents(sb, contents);
	    } else {
	      outputContents(sb, reposition(formula, rowDelta, rowCursor, colDelta, colCursor));
	    }
	    sb.append((j != lastCol1) ? ("\",") : ("\"\n"));
	  }  
	}
	for (int j = colCursor+(-colDelta); j < lastCol; j++) {
	  SheetTableCell td = (SheetTableCell)getTableCell(i, j);
	  String contents = td.getContents();
	  if (contents == null || contents.length() == 0) {
	    sb.append((j != lastCol1) ? (",") : ("\n"));
	  } else {
	    sb.append("\"");
	    String formula = td.getFormula();
	    if (formula == null) { 
	      outputContents(sb, contents);
	    } else {
	      outputContents(sb, reposition(formula, rowDelta, rowCursor, colDelta, colCursor));
	    }
	    sb.append((j != lastCol1) ? ("\",") : ("\"\n"));
	  }  
	}
      }
    } else if (colDelta > 0) {
      DEBUG.p("insert columns " + colDelta);
      // insert columns
      int lastCol2 = lastCol1+colDelta;
      for (int i = 1; i < lastRow; i++) {
	for (int j = 1; j <= insertAtCol; j++) {
	  SheetTableCell td = (SheetTableCell)getTableCell(i, j);
	  String contents = td.getContents();
	  if (contents == null || contents.length() == 0) {
	    sb.append((j != lastCol2) ? (",") : ("\n"));
	  } else {
	    sb.append("\"");
	    String formula = td.getFormula();
	    if (formula == null) { 
	      outputContents(sb, contents);
	    } else {
	      outputContents(sb, reposition(formula, rowDelta, rowCursor, colDelta, insertAtCol));
	    }
	    sb.append((j != lastCol2) ? ("\",") : ("\"\n"));
	  }
	}
	for (int j = insertAtCol+1; j <= insertAtCol+colDelta; j++) {
	  sb.append((j != lastCol2) ? (",") : ("\n"));
	}
	for (int j = insertAtCol+1; j < lastCol; j++) {
	  SheetTableCell td = (SheetTableCell)getTableCell(i, j);
	  String contents = td.getContents();
	  if (contents == null || contents.length() == 0) {
	    sb.append((j != lastCol1) ? (",") : ("\n"));
	  } else {
	    sb.append("\"");
	    String formula = td.getFormula();
	    if (formula == null) { 
	      outputContents(sb, contents);
	    } else {
	      outputContents(sb, reposition(formula, rowDelta, rowCursor, colDelta, insertAtCol));
	    }
	    sb.append((j != lastCol1) ? ("\",") : ("\"\n"));
	  }  
	}
      }
    } else {
      throw new RuntimeException("can't happen");
    }
    String v = sb.toString();
    DEBUG.p("v.length="+v.length());
    return v;
  }

    // Structured comment
    // # !A1 Title
    // or
    // # -A1 Title
  private void outputTitle(StringBuffer sb) {
    {
      sb.append("# ");
      sb.append((isWheelAxisHorizontal()) ? "-" : "|");
      sb.append(getCellName());
      sb.append(" ");
      sb.append(getSheetName());
      sb.append("\n");
    }
  }

  // Structured comment
  // ## 64 12 55 55 55 60
  // means
  // 64 rows (including row 0 for heading)
  // and the column pixels widths are 12 (for heading)
  // and 55,55,55,60.
  // Sum of column widths must be 240-3.
  // Outputs nothing if cell sizes are default.
  private void outputCellWidths(StringBuffer sb, int rowDelta, int colDelta, boolean insertOnLeft, boolean makeTableWider) {
    //DEBUG.p("colCursor="+colCursor+" outputCellWidths(rowDelta="+rowDelta+", "+
    // "colDelta="+colDelta+", "+
    // "insertOnLeft="+insertOnLeft+", "+
    // "makeTableWider="+makeTableWider+")");
    if (cellWidths != DEFAULT_CELL_WIDTHS || rowDelta != 0 || colDelta != 0) {
      sb.append("## ");
      sb.append(nRows + rowDelta);
      sb.append(" ");
      sb.append(cellWidths[0]);
      sb.append(" ");
      if (colDelta == 0) {
	for (int i = 1; i < nCols; i++) {
	  sb.append(cellWidths[i]);
	  sb.append(" ");
	}
      } 
      else if (colDelta < 0) {
	int d = -colDelta;
	if (colCursor == 1) {
	  int w = 0;
	  for (int i = 1; i <= colCursor+d; i++) {
	    w += cellWidths[i];
	  }
	  sb.append(w);
	  sb.append(" ");	  
	  for (int i = colCursor+d+1; i < nCols; i++) {
	    sb.append(cellWidths[i]);
	    sb.append(" ");
	  }
	} else {
	  for (int i = 1; i <= colCursor-1; i++) {
	    sb.append(cellWidths[i]);
	    sb.append(" ");
	  }
	  int w = 0;
	  for (int i = colCursor; i <= colCursor+d; i++) {
	    if (i > 0) {
	      w += cellWidths[i];
	    }
	  }
	  sb.append(w);
	  sb.append(" ");	  
	  for (int i = colCursor+d+1; i < nCols; i++) {
	    sb.append(cellWidths[i]);
	    sb.append(" ");
	  }
	}
      } else {
	  // colDelta > 0
	  int insertAtCol = insertOnLeft ? colCursor-1 : colCursor;
	  if (insertAtCol == 0) insertAtCol = 1; // can't split col 0
	  for (int i = 1; i < nCols; i++) {
	    if (i == insertAtCol) {
	      if (makeTableWider) {
		// Keep column widths -- table will be wider than screen
		int k = cellWidths[i];
		sb.append(k);
		sb.append(" ");
		sb.append(k);
	      } else {
		// Split the current column into two.
		int k = cellWidths[i] / 2;
		sb.append(k);
		sb.append(" ");
		sb.append(cellWidths[i]-k);
	      }
	    } else {
	      sb.append(cellWidths[i]);
	    }
	    sb.append(" ");
	  }
	}
      sb.append("\n");
    }
  }

  // Structured Comment
  // #.0 
  // to 
  // #.9
  // or more!
  private void outputPrecision(StringBuffer sb) {
    sb.append("#.");
    sb.append(getParser().getPrecision());
    sb.append("\n");
  }


  // Structured comments for onLoad actions.
  // #=expr1
  // #=expr2
  private void outputOnLoad(StringBuffer sb, int rowDelta, int colDelta, boolean insertOnLeft) {
    if (onLoad != null) {
      int insertAtCol = insertOnLeft ? colCursor-1 : colCursor;
      for (ListIterator it = onLoad.listIterator(); it.hasNext(); ) {
	String expr = (String)it.next();
	sb.append("#=");
	String formula = reposition(expr, rowDelta, rowCursor, colDelta, insertAtCol);
	if (formula.startsWith("="))
	  formula = formula.substring(1);
	sb.append(formula);
	sb.append("\n");
      }
    }
  }


  private void outputContents(StringBuffer sb, String contents) {
    if (contents.indexOf('"') == -1) {
      sb.append(contents);
    } else {
      int slen = contents.length();
      for (int si = 0; si<slen; si++) {
	char c = contents.charAt(si);
	if (c == '"') sb.append('\\');
	sb.append(c);
      }
    } 
  }

  /**
   * @param expr The expression to be rewritten
   * @param rowDelta positive rows to insert atRow, or negative rows to discard atRow
   * @param atRow The first row to be inserted or discarded.
   * Return a new expr that is updated for the introduction or deletion
   * of rowDelta (or -rowDelta) rows atRow.
   */
  String reposition(String expr, int rowDelta, int atRow, int colDelta, int atCol) {
    if (rowDelta == 0 && colDelta == 0) return expr;
    String result = getParser().reposition(expr, rowDelta, atRow, colDelta, atCol);
    DEBUG.p("reposition("+expr+", rowDelta="+rowDelta+" atRow="+atRow+" colDelta="+colDelta+" atCol="+atCol+") => " + result);
    return result;
  }


  final Object evaluateToObject(String f) {
    Object value;
    try {
      value = getParser().toplevelParse(f).eval();

    } catch (Exception e) {
      SheetWindow.errorprint(e);
      value = e.getMessage();
    }
    return value;
  }

  final Pratt getParser() {
    if (parser == null)
      parser = new Pratt(this);
    return parser;
  }    

  final void setPrecision(int precision) {
    getParser().setPrecision(precision);
  }

  final int getPrecision() {
    return getParser().getPrecision();
  }

  LinkedList calculateDependencies(String f) {
    String value;
    try {
      return getParser().toplevelParse(f).getCellRefs();
    } catch (ParserException e) {
      // If we got a parse error calculating dependencies
      // then return null.  The cell will either have its own error
      // later or we'll have to recalculate this cell anyway
      // when the user edits it.  Otherwise we get multiplie
      // copies of the same error.
      return null;
    } catch (Exception e) {
      // Otherwise let the user know right away.
      SheetWindow.errorprint(e);
      return null;
    }
  }

  /**
   * @returns true if succesful, false if not.
   * @param pixels number of +/- pixels
   * @param left if true, steal pixels from left; else steal from right.
   * @returns false if couldn't steal pixels from specified direction, 
   * or if column would be too small.
   */
  boolean adjustColumnWidth(int pixels, boolean left) {
    if ((left && colCursor == 1) || (!left && colCursor == nCols-1))
      return false;
    if ((cellWidths[colCursor] + pixels) < 1)
      return false;
    int peterCol = (left ? (colCursor - 1) : (colCursor + 1));
    if ((cellWidths[peterCol] - pixels) < 1)
      return false;
    if (cellWidths == DEFAULT_CELL_WIDTHS) 
      cellWidths = ((int[])(cellWidths.clone()));
    cellWidths[peterCol] -= pixels;
    cellWidths[colCursor] += pixels;
    return true;
  }


  /**
   * @returns true if succesful, false if not.
   */
  boolean adjustFitColumnToCell() {
    int stringWidth = getCellStringWidth();
    int columnWidth = cellWidths[colCursor];
    if (stringWidth < columnWidth) {
      return adjustColumnWidth(stringWidth - columnWidth, false);
    } else if (stringWidth > columnWidth) {
      return adjustColumnWidth(stringWidth - columnWidth, false);
    } else 
      return false;
  }

  private static flipsheet getFlipsheet() {
    return ((flipsheet)Application.getCurrentApp());
  }

//  public boolean eventWidgetUp(int i, Event event) {
//    DEBUG.p("SheetTable eventWidgetUp " + i);
//    event.debugPrint();
//    return super.eventWidgetUp(i, event);
//  }
}


